"""Raycast"""

from __future__ import annotations

import chex
import jax
import jax.numpy as jnp

from phyjax2d.impl import (
    Circle,
    Polygon,
    Segment,
    State,
    _sv_cross,
    _vmap_dot,
    normalize,
)


@chex.dataclass
class Raycast:
    fraction: jax.Array
    normal: jax.Array
    hit: jax.Array


def circle_raycast(
    radius: float | jax.Array,
    max_fraction: float | jax.Array,
    p1: jax.Array,
    p2: jax.Array,
    circle: Circle,
    state: State,
) -> Raycast:
    # Suppose p1 and p2's shape has (2,)
    s = jnp.expand_dims(p1, axis=0) - state.p.xy  # (N, 2)
    d, length = normalize(p2 - p1)
    t = -jnp.dot(s, d)  # (N,)

    @jax.vmap
    def muld(x: jax.Array) -> jax.Array:
        return x * d

    c = s + muld(t)  # (N, 2)
    cc = _vmap_dot(c, c)  # (N, 1)
    rr = (radius + circle.radius) ** 2
    fraction = jnp.where(rr >= cc, t - jnp.sqrt(rr - cc), 0.0)
    hitpoint = s + muld(fraction)
    normal, _ = normalize(hitpoint)
    return Raycast(  # type: ignore
        fraction=fraction / length,
        normal=normal,
        hit=jnp.logical_and(
            cc <= rr,
            jnp.logical_and(0.0 <= fraction, fraction <= max_fraction * length),
        ),
    )


def segment_raycast(
    max_fraction: float | jax.Array,
    p1: jax.Array,
    p2: jax.Array,
    segment: Segment,
    state: State,
) -> Raycast:
    d = p2 - p1
    v1, v2 = segment.point1, segment.point2
    v1, v2 = state.p.transform(v1), state.p.transform(v2)
    e = v2 - v1
    eunit, length = normalize(e)
    normal = _sv_cross(jnp.ones_like(length) * -1, eunit)
    numerator = _vmap_dot(normal, v1 - p1)  # (N,)
    denominator = jnp.dot(normal, d)  # (N,)
    t = numerator / denominator
    p = jax.vmap(lambda ti: ti * d + p1)(t)  # (N, 2)
    s = _vmap_dot(p - v1, eunit)
    normal = jnp.where(jnp.expand_dims(numerator > 0.0, axis=1), -normal, normal)
    return Raycast(  # type: ignore
        fraction=t,
        normal=normal,
        hit=jnp.logical_and(
            denominator != 0.0,
            jnp.logical_and(
                jnp.logical_and(t >= 0.0, t <= max_fraction),
                jnp.logical_and(s >= 0.0, s <= length),
            ),
        ),
    )


def thin_polygon_raycast(
    max_fraction: float | jax.Array,
    p1: jax.Array,
    p2: jax.Array,
    polygon: Polygon,
    state: State,
) -> Raycast:
    p1 = state.p.transform(p1)  # (N, 2)
    d = state.p.transform(p2) - p1  # (N, 2)
    vp = polygon.points - jnp.expand_dims(p1, axis=1)  # (N, NP, 2)
    numerator = _vmap_dot(polygon.normals, vp)  # (N, NP)
    denominator = jnp.einsum("npt,nt->np", polygon.normals, d)
    t = numerator / denominator
    upper = jnp.min(jnp.where(denominator > 0.0, t, jnp.inf), axis=1)
    lower_cand = jnp.where(denominator < 0.0, t, jnp.inf)
    lower = jnp.min(lower_cand, axis=1)
    idx = jnp.argmin(lower_cand, axis=1)
    return Raycast(  # type: ignore
        fraction=lower,
        normal=polygon.normals[jnp.arange(idx.shape[0]), idx.ravel()],
        hit=jnp.logical_and(
            jnp.logical_and(lower >= 0.0, lower <= max_fraction),
            lower < upper,
        ),
    )



"""Utilities to construct physics simulation"""

from __future__ import annotations

import dataclasses
import warnings
from collections import defaultdict
from copy import deepcopy
from typing import Any, Callable, NamedTuple, TypeVar

import jax
import jax.numpy as jnp
import numpy as np

from phyjax2d.impl import (
    Capsule,
    Circle,
    Polygon,
    Segment,
    Shape,
    ShapeDict,
    Space,
    State,
    StateDict,
    _vmap_dot,
    empty,
)
from phyjax2d.vec2d import Vec2d

Self = Any
_N_MAX_POLYGON_VERTICES = 6

_ALL_POLYGON_KEYS = [
    "triangle",
    "static_triangle",
    "quadrangle",
    "static_quadrangle",
    "pentagon",
    "static_pentagon",
    "hexagon",
    "static_hexagon",
]

_POLYGON_NAMES = {3: "triangle", 4: "quadrangle", 5: "pentagon", 6: "hexagon"}


class Color(NamedTuple):
    r: int
    g: int
    b: int
    a: int = 255

    @staticmethod
    def from_float(r: float, g: float, b: float, a: float = 1.0) -> Self:
        return Color(int(r * 255), int(g * 255), int(b * 255), int(a * 255))

    @staticmethod
    def black() -> Self:
        return Color(0, 0, 0, 255)


_BLACK = Color.black()


def _mass_and_moment(
    mass: jax.Array,
    moment: jax.Array,
    is_static: bool = False,
) -> tuple[jax.Array, jax.Array]:
    if is_static:
        return jnp.array(jnp.inf), jnp.array(jnp.inf)
    else:
        return jnp.array(mass), jnp.array(moment)


def _circle_mass(radius: float, density: float) -> tuple[jax.Array, jax.Array]:
    rr = radius**2
    mass = density * jnp.pi * rr
    moment = 0.5 * mass * rr
    return jnp.array(mass), jnp.array(moment)


def _capsule_mass(
    radius: float,
    length: float,
    density: float,
) -> tuple[jax.Array, jax.Array]:
    rr, ll = radius**2, length**2
    mass = density * (jnp.pi * radius + 2.0 * length) * radius
    circle_moment = 0.5 * (rr + ll)
    box_moment = (4 * rr + ll) / 12
    moment = mass * (circle_moment + box_moment)
    return jnp.array(mass), jnp.array(moment)


def _polygon_mass(
    points: list[Vec2d],
    normals: list[Vec2d],
    radius: float,
    density: float,
) -> tuple[jax.Array, jax.Array]:
    vertices = deepcopy(points)
    n = len(vertices)
    # If the polygon is rounded, approximate mass by pushing out vertices
    if radius > 0:
        radius_sqrt2 = radius * (2**0.5)
        for i in range(n):
            j = (i + 1) % n
            n1 = normals[i]
            n2 = normals[j]
            mid = (n1 + n2).normalized()
            vertices[i] += radius_sqrt2 * mid

    r = vertices[0]
    total_area = 0.0
    center = Vec2d(0.0, 0.0)
    moment = 0.0
    for v1, v2 in zip(vertices[1:], vertices[2:]):
        e1 = v1 - r
        e2 = v2 - r
        d = e1.cross(e2)
        area = d * 0.5
        total_area += area
        center += (e1 + e2) * (area / 3)
        intx2 = e1.x**2 + e2.x**2 + e1.x * e2.x
        inty2 = e1.y**2 + e2.y**2 + e1.y * e2.y
        moment += d * (intx2 + inty2) / 12

    mass = density * total_area
    center /= total_area
    center_r = center + r
    moment_shift = center_r.dot(center_r) - center.dot(center)
    moment = (density * moment) + mass * moment_shift
    return jnp.array(mass), jnp.array(moment)


S = TypeVar("S", bound=Shape)


def _stack_or(sl: list[S], default_fn: Callable[[], S]) -> S:
    if len(sl) > 0:
        return jax.tree.map(lambda *args: jnp.stack(args), *sl)
    else:
        return default_fn()


def _check_params_positive(friction: float, **kwargs) -> None:
    if friction > 1.0:
        warnings.warn(
            f"friction larger than 1 can lead instable simulation (value: {friction})",
            stacklevel=2,
        )
    for key, value in kwargs.items():
        assert value > 0.0, f"Invalid value for {key}: {value}"


def _compute_centroid(points: list[Vec2d]) -> Vec2d:
    # Compute centroid
    center = Vec2d(0.0, 0.0)
    total_area = 0.0
    origin = points[0]
    for a, b in zip(points[1:], points[2:]):
        e1 = a - origin
        e2 = b - origin
        area = e1.cross(e2) * 0.5
        center += (e1 + e2) * (area / 3)
        total_area += area
    return (center / total_area) + origin


def _make_listdd() -> dict[int, list]:
    return defaultdict(list)


@dataclasses.dataclass
class SpaceBuilder:
    """
    A convenient builder for creating a space.
    Not expected to used with `jax.jit`.
    """

    gravity: Vec2d | tuple[float, float] = dataclasses.field(default=(0.0, -9.8))
    circles: list[Circle] = dataclasses.field(default_factory=list)
    static_circles: list[Circle] = dataclasses.field(default_factory=list)
    capsules: list[Capsule] = dataclasses.field(default_factory=list)
    static_capsules: list[Capsule] = dataclasses.field(default_factory=list)
    segments: list[Segment] = dataclasses.field(default_factory=list)
    polygons: dict[int, list[Polygon]] = dataclasses.field(default_factory=_make_listdd)
    static_polygons: dict[int, list[Polygon]] = dataclasses.field(
        default_factory=_make_listdd
    )
    dt: float = 0.1
    linear_damping: float = 0.9
    angular_damping: float = 0.9
    bias_factor: float = 0.2
    n_velocity_iter: int = 6
    n_position_iter: int = 2
    linear_slop: float = 0.005
    max_linear_correction: float = 0.2
    allowed_penetration: float = 0.005
    bounce_threshold: float = 1.0
    max_velocity: float | None = None
    max_angular_velocity: float | None = None
    _ignore_constraints: dict[tuple[str, str], list[int]] = dataclasses.field(
        default_factory=lambda: defaultdict(list),
        init=False,
    )

    def add_circle(
        self,
        *,
        radius: float,
        density: float = 1.0,
        is_static: bool = False,
        friction: float = 0.8,
        elasticity: float = 0.8,
        ignore: list[str] | None = None,
        color: Color = _BLACK,
    ) -> None:
        _check_params_positive(
            friction=friction,
            radius=radius,
            density=density,
            elasticity=elasticity,
        )
        mass, moment = _mass_and_moment(*_circle_mass(radius, density), is_static)
        circle = Circle(
            radius=jnp.array(radius),
            mass=mass,
            moment=moment,
            elasticity=jnp.array(elasticity),
            friction=jnp.array(friction),
            rgba=jnp.array(color),
        )
        if is_static:
            self._add_ignore_constraint(
                this="static_circle",
                index=len(self.static_circles),
                ignore=ignore,
            )
            self.static_circles.append(circle)
        else:
            self._add_ignore_constraint(
                this="circle",
                index=len(self.circles),
                ignore=ignore,
            )
            self.circles.append(circle)

    def add_capsule(
        self,
        *,
        p1: Vec2d,
        p2: Vec2d,
        radius: float,
        density: float = 1.0,
        is_static: bool = False,
        friction: float = 0.8,
        elasticity: float = 0.8,
        ignore: list[str] | None = None,
        color: Color = _BLACK,
    ) -> None:
        _check_params_positive(
            friction=friction,
            radius=radius,
            density=density,
            elasticity=elasticity,
        )
        mass, moment = _mass_and_moment(
            *_capsule_mass(radius, (p2 - p1).length, density),
            is_static,
        )
        capsule = Capsule(
            point1=jnp.array(p1),
            point2=jnp.array(p2),
            radius=jnp.array(radius),
            mass=mass,
            moment=moment,
            elasticity=jnp.array(elasticity),
            friction=jnp.array(friction),
            rgba=jnp.array(color),
        )
        if is_static:
            self._add_ignore_constraint(
                this="static_capsule",
                index=len(self.static_capsules),
                ignore=ignore,
            )
            self.static_capsules.append(capsule)
        else:
            self._add_ignore_constraint(
                this="capsule",
                index=len(self.capsules),
                ignore=ignore,
            )
            self.capsules.append(capsule)

    def add_segment(
        self,
        *,
        p1: Vec2d,
        p2: Vec2d,
        friction: float = 0.8,
        elasticity: float = 0.8,
        ignore: list[str] | None = None,
        color: Color = _BLACK,
    ) -> None:
        _check_params_positive(
            friction=friction,
            elasticity=elasticity,
        )
        point1 = jnp.array(p1)
        point2 = jnp.array(p2)
        segment = Segment(
            point1=point1,
            point2=point2,
            is_smooth=jnp.array(False),
            # Fake ghosts
            ghost1=point1,
            ghost2=point2,
            mass=jnp.array(jnp.inf),
            moment=jnp.array(jnp.inf),
            elasticity=jnp.array(elasticity),
            friction=jnp.array(friction),
            rgba=jnp.array(color),
        )
        self._add_ignore_constraint(
            this="segment",
            index=len(self.segments),
            ignore=ignore,
        )
        self.segments.append(segment)

    def add_polygon(
        self,
        *,
        points: list[Vec2d],
        radius: float = 0.0,
        friction: float = 0.8,
        elasticity: float = 0.8,
        density: float = 1.0,
        is_static: bool = False,
        ignore: list[str] | None = None,
        color: Color = _BLACK,
    ) -> None:
        _check_params_positive(
            friction=friction,
            elasticity=elasticity,
        )
        # Need to have at least three points
        n = len(points)
        if n < 3:
            raise ValueError("Polygon needs at least three vertices")
        elif n > _N_MAX_POLYGON_VERTICES:
            raise ValueError(f"Too many vertices {n} for polygon")
        # Check convexity
        cross_list = []
        for a, b, c in zip(points, points[1:], points[2:]):
            a2b = b - a
            b2c = c - b
            cross = a2b.cross(b2c)
            if cross == 0:
                raise ValueError(f"Redundant points in polygon: {a, b, c}")
            cross_list.append(cross)
        signs = np.sign(cross_list)
        if np.all(signs == -1):
            # Flip arrays to make them counter clock wise
            points.reverse()
        elif not np.all(signs == 1):
            raise ValueError("Given polygon is not convex!")
        self._add_polygon_internal(
            points=points,
            centroid=_compute_centroid(points),
            radius=radius,
            friction=friction,
            elasticity=elasticity,
            density=density,
            is_static=is_static,
            ignore=ignore,
            rgba=color,
        )

    def add_square(
        self,
        *,
        width: float,
        height: float,
        radius: float = 0.0,
        friction: float = 0.8,
        elasticity: float = 0.8,
        density: float = 1.0,
        is_static: bool = False,
        ignore: list[str] | None = None,
        color: Color = _BLACK,
    ) -> None:
        a = Vec2d(width / 2, height / 2)
        b = Vec2d(-width / 2, height / 2)
        c = Vec2d(-width / 2, -height / 2)
        d = Vec2d(width / 2, -height / 2)
        self._add_polygon_internal(
            points=[a, b, c, d],
            centroid=Vec2d(0.0, 0.0),
            radius=radius,
            friction=friction,
            elasticity=elasticity,
            density=density,
            is_static=is_static,
            ignore=ignore,
            rgba=color,
        )

    def _add_polygon_internal(
        self,
        *,
        points: list[Vec2d],
        centroid: Vec2d,
        radius: float = 0.0,
        friction: float = 0.8,
        elasticity: float = 0.8,
        density: float = 1.0,
        is_static: bool = False,
        ignore: list[str] | None = None,
        rgba: Color = _BLACK,
    ) -> None:
        # Compute normal
        normals = []

        for p1, p2 in zip(points, points[1:] + points[:1]):
            edge = p2 - p1
            if edge.dot(edge) < 1e-6:
                raise ValueError(f"Edge is too short in polygon: {p1} and {p2}")
            # Rotate the edge 90 degree right and normalize it
            normals.append(edge.perpendicular_right().normalized())

        mass, moment = _polygon_mass(points, normals, radius, density)
        polygon = Polygon(
            points=jnp.array(points),
            normals=jnp.array(normals),
            centroid=jnp.array(centroid),
            radius=jnp.array(radius),
            mass=mass,
            moment=moment,
            elasticity=jnp.array(elasticity),
            friction=jnp.array(friction),
            rgba=jnp.array(rgba),
        )
        n = len(points)
        if is_static:
            self._add_ignore_constraint(
                this="static_" + _POLYGON_NAMES[n],
                index=len(self.polygons[n]),
                ignore=ignore,
            )
            self.static_polygons[n].append(polygon)
        else:
            self._add_ignore_constraint(
                this=_POLYGON_NAMES[n],
                index=len(self.polygons[n]),
                ignore=ignore,
            )
            self.polygons[n].append(polygon)

    def _add_ignore_constraint(
        self, this: str, index: int, ignore: list[str] | None = None
    ) -> None:
        if ignore is not None:
            for ignore_name in ignore:
                self._ignore_constraints[this, ignore_name].append(index)

    def add_chain_segments(
        self,
        *,
        chain_points: list[tuple[Vec2d, Vec2d]],
        friction: float = 0.8,
        elasticity: float = 0.8,
        color: Color = _BLACK,
    ) -> None:
        _check_params_positive(
            friction=friction,
            elasticity=elasticity,
        )
        n_points = len(chain_points)
        for i in range(n_points):
            g1 = chain_points[i - 1][0]
            p1, p2 = chain_points[i]
            g2 = chain_points[(i + 1) % n_points][1]
            segment = Segment(
                point1=jnp.array(p1),
                point2=jnp.array(p2),
                is_smooth=jnp.array(True),
                # Fake ghosts
                ghost1=jnp.array(g1),
                ghost2=jnp.array(g2),
                mass=jnp.array(jnp.inf),
                moment=jnp.array(jnp.inf),
                elasticity=jnp.array(elasticity),
                friction=jnp.array(friction),
                rgba=jnp.array(color),
            )
            self.segments.append(segment)

    def build(self) -> Space:
        shaped = ShapeDict(
            circle=_stack_or(self.circles, empty(Circle)),
            static_circle=_stack_or(self.static_circles, empty(Circle)),
            segment=_stack_or(self.segments, empty(Segment)),
            capsule=_stack_or(self.capsules, empty(Capsule)),
            static_capsule=_stack_or(self.static_capsules, empty(Capsule)),
            triangle=_stack_or(self.polygons[3], empty(Polygon)),
            static_triangle=_stack_or(self.static_polygons[3], empty(Polygon)),
            quadrangle=_stack_or(self.polygons[4], empty(Polygon)),
            static_quadrangle=_stack_or(self.static_polygons[4], empty(Polygon)),
            pentagon=_stack_or(self.polygons[5], empty(Polygon)),
            static_pentagon=_stack_or(self.static_polygons[5], empty(Polygon)),
            hexagon=_stack_or(self.polygons[6], empty(Polygon)),
            static_hexagon=_stack_or(self.static_polygons[6], empty(Polygon)),
        )
        linear_damping = jnp.exp(-self.dt * self.linear_damping).item()
        angular_damping = jnp.exp(-self.dt * self.angular_damping).item()
        max_velocity = jnp.inf if self.max_velocity is None else self.max_velocity
        max_angular_velocity = (
            jnp.inf if self.max_angular_velocity is None else self.max_angular_velocity
        )
        return Space(
            gravity=jnp.array(self.gravity),
            shaped=shaped,
            linear_damping=linear_damping,
            angular_damping=angular_damping,
            bias_factor=self.bias_factor,
            n_velocity_iter=self.n_velocity_iter,
            n_position_iter=self.n_position_iter,
            linear_slop=self.linear_slop,
            max_linear_correction=self.max_linear_correction,
            allowed_penetration=self.allowed_penetration,
            bounce_threshold=self.bounce_threshold,
            max_velocity=max_velocity,
            max_angular_velocity=max_angular_velocity,
            ignore_collision=self._ignore_constraints,
        )


def make_approx_circle(
    center: Vec2d | tuple[float, float],
    radius: float,
    n_lines: int = 32,
) -> list[tuple[Vec2d, Vec2d]]:
    """Make circle. Points are ordered clockwith."""
    unit = np.pi * 2 / n_lines
    lines = []
    t0 = Vec2d(radius, 0.0)
    for i in reversed(range(n_lines)):
        start = center + t0.rotated(unit * (i + 1))
        end = center + t0.rotated(unit * i)
        lines.append((start, end))
    return lines


def make_square_segments(
    xmin: float,
    xmax: float,
    ymin: float,
    ymax: float,
    rounded_offset: float | None = None,
) -> list[tuple[Vec2d, Vec2d]]:
    """Make square. Points are ordered clockwith."""
    p1 = Vec2d(xmin, ymin)
    p2 = Vec2d(xmin, ymax)
    p3 = Vec2d(xmax, ymax)
    p4 = Vec2d(xmax, ymin)
    lines = []
    if rounded_offset is not None:
        for start, end in [(p1, p2), (p2, p3), (p3, p4), (p4, p1)]:
            s2end = Vec2d(*end) - Vec2d(*start)
            offset = s2end.normalized() * rounded_offset
            stop = end - offset
            lines.append((start + offset, stop))
            # Center of the rounded corner
            center = stop + offset.rotated(-np.pi / 2)
            for i in reversed(range(4)):
                r_start = center + offset.rotated(np.pi / 8 * (i + 1))
                r_end = center + offset.rotated(np.pi / 8 * i)
                lines.append((r_start, r_end))
    else:
        for start, end in [(p1, p2), (p2, p3), (p3, p4), (p4, p1)]:
            lines.append((start, end))
    return lines


def _circle_polygon_overlap(
    polygon: Polygon,
    pstate: State,
    xy: jax.Array,
    radius: jax.Array | float,
) -> jax.Array:
    n_batch, n_vertices = polygon.points.shape[:2]
    # Suppose that pstate.p.xy.shape == (N, 2) and xy.shape == (2,)
    cxy = pstate.p.inv_transform(jnp.expand_dims(xy, axis=0))
    p2cxy = jnp.expand_dims(cxy, axis=1) - polygon.points
    separation = _vmap_dot(polygon.normals, p2cxy)  # (N, NP)
    max_sep = jnp.max(separation, axis=1)
    i1 = jnp.argmax(separation, axis=1)
    i2 = (i1 + 1) % n_vertices
    select_all = jnp.arange(n_batch)
    v1 = polygon.points[select_all, i1]
    v2 = polygon.points[select_all, i2]
    u1 = _vmap_dot(cxy - v1, v2 - v1)
    u2 = _vmap_dot(cxy - v2, v1 - v2)
    v = jnp.where(jnp.expand_dims(u1 < 0.0, axis=1), v1, v2)
    dist = jnp.linalg.norm(cxy - v, axis=1)
    c_out = dist < polygon.radius + radius
    c_in = max_sep < polygon.radius + radius
    return jax.lax.select(jnp.logical_or(u1 < 0.0, u2 < 0.0), c_out, c_in)


def circle_overlap(
    shaped: ShapeDict,
    stated: StateDict,
    xy: jax.Array,
    radius: jax.Array | float,
) -> jax.Array:
    # Circle overlap
    overlap = jnp.array(False)
    if not stated.circle.is_empty() and not shaped.circle.is_empty():
        cpos = stated.circle.p.xy
        # Suppose that cpos.shape == (N, 2) and xy.shape == (2,)
        dist = jnp.linalg.norm(cpos - jnp.expand_dims(xy, axis=0), axis=-1)
        penetration = shaped.circle.radius + radius - dist
        has_overlap = jnp.logical_and(stated.circle.is_active, penetration >= 0)
        overlap = jnp.any(has_overlap)

    # Static_circle overlap
    if not stated.static_circle.is_empty() and not shaped.static_circle.is_empty():
        cpos = stated.static_circle.p.xy
        # Suppose that cpos.shape == (N, 2) and xy.shape == (2,)
        dist = jnp.linalg.norm(cpos - jnp.expand_dims(xy, axis=0), axis=-1)
        penetration = shaped.static_circle.radius + radius - dist
        has_overlap = jnp.logical_and(stated.static_circle.is_active, penetration >= 0)
        overlap = jnp.logical_or(jnp.any(has_overlap), overlap)

    # Circle-segment overlap
    if not stated.segment.is_empty() and not shaped.segment.is_empty():
        spos = stated.segment.p
        # Suppose that spos.shape == (N, 2) and xy.shape == (2,)
        pb = spos.inv_transform(jnp.expand_dims(xy, axis=0))
        p1, p2 = shaped.segment.point1, shaped.segment.point2
        edge = p2 - p1
        s1 = jnp.expand_dims(_vmap_dot(pb - p1, edge), axis=1)
        s2 = jnp.expand_dims(_vmap_dot(p2 - pb, edge), axis=1)
        in_segment = jnp.logical_and(s1 >= 0.0, s2 >= 0.0)
        ee = jnp.sum(jnp.square(edge), axis=-1, keepdims=True)
        pa = jnp.where(in_segment, p1 + edge * s1 / ee, jnp.where(s1 < 0.0, p1, p2))
        dist = jnp.linalg.norm(pb - pa, axis=-1)
        penetration = radius - dist
        has_overlap = jnp.logical_and(stated.segment.is_active, penetration >= 0)
        overlap = jnp.logical_or(jnp.any(has_overlap), overlap)

    # Circle-polygon overlap
    for key in _ALL_POLYGON_KEYS:
        if not stated[key].is_empty() and not shaped[key].is_empty():  # type: ignore
            has_overlap = _circle_polygon_overlap(
                shaped[key],  # type: ignore
                stated[key],  # type: ignore
                xy,
                radius,
            )
            overlap = jnp.logical_or(jnp.any(has_overlap), overlap)

    return overlap



"""
A simple and fast visualizer based on moderngl.
Currently, only supports circles and lines.
"""

from __future__ import annotations

from collections.abc import Callable
from typing import ClassVar

import jax
import jax.numpy as jnp
import moderngl as mgl
import moderngl_window as mglw
import numpy as np
from moderngl_window.context import headless
from numpy.typing import NDArray

from phyjax2d import Circle, Polygon, Position, Segment, Space, State, StateDict

NOWHERE: float = -1000.0


_CIRCLE_VERTEX_SHADER = """
#version 330
uniform mat4 proj;
in vec2 in_position;
in float in_scale;
in vec4 in_color;
out vec4 v_color;
void main() {
    gl_Position = proj * vec4(in_position, 0.0, 1.0);
    gl_PointSize = in_scale;
    v_color = in_color;
}
"""

# Smoothing by fwidth is based on: https://rubendv.be/posts/fwidth/
_CIRCLE_FRAGMENT_SHADER = """
#version 330
in vec4 v_color;
out vec4 f_color;
void main() {
    float dist = length(gl_PointCoord.xy - vec2(0.5));
    float delta = fwidth(dist);
    float alpha = smoothstep(0.5, 0.5 - delta, dist);
    f_color = v_color * alpha;
}
"""

_LINE_VERTEX_SHADER = """
#version 330
in vec2 in_position;
uniform mat4 proj;
void main() {
    gl_Position = proj * vec4(in_position, 0.0, 1.0);
}
"""

_LINE_GEOMETRY_SHADER = """
#version 330
layout (lines) in;
layout (triangle_strip, max_vertices = 4) out;
uniform float width;
out float g_len;
out float g_u;
out float g_v;
void main() {
    vec2 a = gl_in[0].gl_Position.xy;
    vec2 b = gl_in[1].gl_Position.xy;
    vec2 a2b = b - a;
    vec2 a2left = vec2(-a2b.y, a2b.x) / length(a2b) * width;
    float len = length(a2b) * 0.5;
    vec4 positions[4] = vec4[4](
        vec4(a + a2left, 0.0, 1.0),
        vec4(a - a2left, 0.0, 1.0),
        vec4(b + a2left, 0.0, 1.0),
        vec4(b - a2left, 0.0, 1.0)
    );
    float gus[4] = float[4](width, -width, width, -width);
    float gvs[4] = float[4](len, len, -len, -len);
    for (int i = 0; i < 4; ++i) {
        g_len = len;
        g_u = gus[i];
        g_v = gvs[i];
        gl_Position = positions[i];
        EmitVertex();
    }
    EmitVertex();
    EndPrimitive();
}
"""

_LINE_FRAGMENT_SHADER = """
#version 330
in float g_u;
in float g_v;
in float g_len;
out vec4 f_color;
uniform float width;
uniform float w_rad;
uniform float l_rad;
uniform vec4 color;
void main() {
    float aw = 1.0 - smoothstep(1.0 - ((2.0 * w_rad) / width), 1.0, abs(g_u / width));
    float al = 1.0 - smoothstep(1.0 - ((2.0 * l_rad) / g_len), 1.0, abs(g_v / g_len));
    f_color = color;
    f_color.a *= min(aw, al);
}
"""

_TRIANGLE_VERTEX_SHADER = """
#version 330
uniform mat4 proj;
in vec2 in_position;
in vec4 in_color;
out vec4 v_color;
void main() {
    gl_Position = proj * vec4(in_position, 0.0, 1.0);
    v_color = in_color;
}
"""


_TRIANGLE_GEOMETRY_SHADER = """
#version 330
layout (triangles) in;
layout (triangle_strip, max_vertices = 3) out;
in vec4 v_color[3];
out vec4 g_color;
void main() {
    for(int i = 0; i < gl_in.length(); i++) {
        gl_Position = gl_in[i].gl_Position;
        g_color = v_color[i];
        EmitVertex();
    }
    EndPrimitive();
}
"""

_TRIANGLE_FRAGMENT_SHADER = """
#version 330
in vec4 g_color;
out vec4 f_color;
void main() {
    f_color = vec4(0.0, 0.0, 0.4, 1.0);
}
"""


class Renderable:
    MODE: ClassVar[int]
    vertex_array: mgl.VertexArray

    def render(self) -> None:
        self.vertex_array.render(mode=self.MODE)


class CircleVA(Renderable):
    MODE = mgl.POINTS

    def __init__(
        self,
        ctx: mgl.Context,
        program: mgl.Program,
        points: NDArray,
        scales: NDArray,
        colors: NDArray,
    ) -> None:
        self._ctx = ctx
        self._length = points.shape[0]
        self._points = ctx.buffer(reserve=len(points) * 4 * 2 * 10)
        self._scales = ctx.buffer(reserve=len(scales) * 4 * 10)
        self._colors = ctx.buffer(reserve=len(colors) * 4 * 4 * 10)

        self.vertex_array = ctx.vertex_array(
            program,
            [
                (self._points, "2f", "in_position"),
                (self._scales, "f", "in_scale"),
                (self._colors, "4f", "in_color"),
            ],
        )
        self.update(points, scales, colors)

    def update(self, points: NDArray, scales: NDArray, colors: NDArray) -> bool:
        length = points.shape[0]
        if self._length != length:
            self._length = length
            self._points.orphan(length * 4 * 2)
            self._scales.orphan(length * 4)
            self._colors.orphan(length * 4 * 4)
        self._points.write(points)
        self._scales.write(scales)
        self._colors.write(colors)
        return length > 0


class SegmentVA(Renderable):
    MODE = mgl.LINES

    def __init__(
        self,
        ctx: mgl.Context,
        program: mgl.Program,
        segments: NDArray,
    ) -> None:
        self._ctx = ctx
        self._length = segments.shape[0]
        self._segments = ctx.buffer(reserve=self._length * 4 * 2 * 10)

        self.vertex_array = ctx.vertex_array(
            program,
            [(self._segments, "2f", "in_position")],
        )
        self.update(segments)

    def update(self, segments: NDArray) -> bool:
        length = segments.shape[0]
        if self._length != length:
            self._length = length
            self._segments.orphan(length * 4 * 2)
        self._segments.write(segments)
        return length > 0


class TriangleVA(Renderable):
    MODE = mgl.TRIANGLES

    def __init__(
        self,
        ctx: mgl.Context,
        program: mgl.Program,
        vertices: NDArray,
        colors: NDArray,
    ) -> None:
        self._ctx = ctx
        self._length = vertices.shape[0]
        self._vertices = ctx.buffer(reserve=self._length * 4 * 6 * 10)
        self._colors = ctx.buffer(reserve=len(colors) * 4 * 12 * 10)

        self.vertex_array = ctx.vertex_array(
            program,
            [
                (self._vertices, "2f", "in_position"),
                # (self._colors, "4f", "in_color"),
            ],
        )
        self.update(vertices)

    def update(self, vertices: NDArray) -> bool:
        length = vertices.shape[0]
        if self._length != length:
            self._length = length
            self._vertices.orphan(length * 4 * 3)
        self._vertices.write(vertices)
        return length > 0


class TextureVA(Renderable):
    MODE = mgl.TRIANGLE_STRIP

    def __init__(
        self,
        ctx: mgl.Context,
        program: mgl.Program,
        texture: mgl.Texture,
    ) -> None:
        self._ctx = ctx
        self._texture = texture
        quad_mat = np.array(
            # x, y, u, v
            [
                [0, 1, 0, 1],  # upper left
                [0, 0, 0, 0],  # lower left
                [1, 1, 1, 1],  # upper right
                [1, 0, 1, 0],  # lower right
            ],
            dtype=np.float32,
        )
        quad_mat_buffer = ctx.buffer(data=quad_mat)
        self.vertex_array = ctx.vertex_array(
            program,
            [(quad_mat_buffer, "2f 2f", "in_position", "in_uv")],
        )

    def update(self, image: bytes) -> None:
        self._texture.write(image)
        self._texture.use()


def _collect_circles(
    circle: Circle,
    state: State,
    circle_scaling: float,
) -> tuple[NDArray, NDArray, NDArray]:
    flag = np.array(state.is_active).reshape(-1, 1)
    points = np.where(flag, np.array(state.p.xy, dtype=np.float32), NOWHERE)
    scales = circle.radius * circle_scaling
    colors = np.array(circle.rgba, dtype=np.float32) / 255.0
    is_active = np.expand_dims(np.array(state.is_active), axis=1)
    colors = np.where(is_active, colors, np.ones_like(colors))
    return points, np.array(scales, dtype=np.float32), colors


@jax.vmap
def vmap_transform(p: Position, points: jax.Array) -> jax.Array:
    return p.transform(points)


def _collect_triangles(
    triangle: Polygon,
    state: State,
) -> tuple[NDArray, NDArray]:
    flag = np.array(state.is_active).reshape(-1, 1, 1)
    points_t = vmap_transform(state.p, triangle.points)
    points = np.where(flag, np.array(points_t), NOWHERE)
    colors = np.array(triangle.rgba, dtype=np.float32) / 255.0
    is_active = np.expand_dims(np.array(state.is_active), axis=1)
    colors = np.where(is_active, colors, np.ones_like(colors))
    return points.reshape(-1, 2), np.repeat(colors, 3, axis=0)


def _collect_static_lines(segment: Segment, state: State) -> NDArray:
    a, b = segment.point1, segment.point2
    a = state.p.transform(a)
    b = state.p.transform(b)
    flag = np.repeat(np.array(state.is_active), 2).reshape(-1, 1)
    return np.where(flag, np.concatenate((a, b), axis=1).reshape(-1, 2), NOWHERE)


def _collect_heads(circle: Circle, state: State) -> NDArray:
    y = jnp.array(circle.radius)
    x = jnp.zeros_like(y)
    p1, p2 = jnp.stack((x, y * 0.8), axis=1), jnp.stack((x, y * 1.2), axis=1)
    p1, p2 = state.p.transform(p1), state.p.transform(p2)
    flag = np.repeat(np.array(state.is_active), 2).reshape(-1, 1)
    return np.where(flag, np.concatenate((p1, p2), axis=1).reshape(-1, 2), NOWHERE)


def _get_clip_ranges(lengthes: list[float]) -> list[tuple[float, float]]:
    """Clip ranges to [-1, 1]"""
    total = sum(lengthes)
    res = []
    left = -1.0
    for length in lengthes:
        right = left + 2.0 * length / total
        res.append((left, right))
        left = right
    return res


def _get_sc_color(colors: NDArray, state: State) -> NDArray:
    # Clip labels to make it work when less number of colors are provided
    label = np.clip(np.array(state.label), 0, len(colors) - 1)
    default_color = colors[label].astype(np.float32) / 255.0
    inactive_color = np.ones_like(default_color)
    is_active_expanded = np.expand_dims(state.is_active, axis=1)
    return np.where(is_active_expanded, default_color, inactive_color)


class MglRenderer:
    """Render pymunk environments to the given moderngl context."""

    def __init__(
        self,
        context: mgl.Context,
        screen_width: int,
        screen_height: int,
        x_range: float,
        y_range: float,
        space: Space,
        stated: StateDict,
        voffsets: tuple[int, ...] = (),
        hoffsets: tuple[int, ...] = (),
        sc_color_opt: NDArray | None = None,
        sensor_color: NDArray | None = None,
        sensor_width: float = 0.001,
        sensor_fn: Callable[[StateDict], tuple[NDArray, NDArray]] | None = None,
    ) -> None:
        self._context = context

        self._screen_x = _get_clip_ranges([screen_width, *hoffsets])
        self._screen_y = _get_clip_ranges([screen_height, *voffsets])
        self._x_range, self._y_range = x_range, y_range
        self._range_min = min(x_range, y_range)

        if x_range < y_range:
            self._range_min = x_range
            self._circle_scaling = screen_width / x_range * 2
        else:
            self._range_min = y_range
            self._circle_scaling = screen_height / y_range * 2

        if sc_color_opt is None:
            self._sc_color = np.array([[254, 2, 162, 255]])
        else:
            self._sc_color = sc_color_opt

        self._space = space
        circle_program = self._make_gl_program(
            vertex_shader=_CIRCLE_VERTEX_SHADER,
            fragment_shader=_CIRCLE_FRAGMENT_SHADER,
        )
        if space.shaped.circle.is_empty():
            self._circles = None
        else:
            points, scales, colors = _collect_circles(
                space.shaped.circle,
                stated.circle,
                self._circle_scaling,
            )
            self._circles = CircleVA(
                ctx=context,
                program=circle_program,
                points=points,
                scales=scales,
                colors=colors,
            )
        if space.shaped.static_circle.is_empty():
            self._static_circles = None
        else:
            points, scales, _ = _collect_circles(
                space.shaped.static_circle,
                stated.static_circle,
                self._circle_scaling,
            )
            self._static_circles = CircleVA(
                ctx=context,
                program=circle_program,
                points=points,
                scales=scales,
                colors=_get_sc_color(self._sc_color, stated.static_circle),
            )
        if space.shaped.segment.is_empty():
            self._static_lines = None
        else:
            static_segment_program = self._make_gl_program(
                vertex_shader=_LINE_VERTEX_SHADER,
                geometry_shader=_LINE_GEOMETRY_SHADER,
                fragment_shader=_LINE_FRAGMENT_SHADER,
                color=np.array([0.0, 0.0, 0.0, 0.4], dtype=np.float32),
                width=np.array([0.004], dtype=np.float32),
                w_rad=np.array([0.001], dtype=np.float32),
                l_rad=np.array([0.001], dtype=np.float32),
            )
            self._static_line_points = _collect_static_lines(
                space.shaped.segment,
                stated.segment,
            )
            self._static_lines = SegmentVA(
                ctx=context,
                program=static_segment_program,
                segments=points,
            )

        if space.shaped.static_triangle.is_empty():
            self._triangles = None
        else:
            points, colors = _collect_triangles(
                space.shaped.static_triangle,
                stated.static_triangle,
            )
            triangle_program = self._make_gl_program(
                vertex_shader=_TRIANGLE_VERTEX_SHADER,
                geometry_shader=_TRIANGLE_GEOMETRY_SHADER,
                fragment_shader=_TRIANGLE_FRAGMENT_SHADER,
            )
            self._triangles = TriangleVA(
                ctx=context,
                program=triangle_program,
                vertices=points,
                colors=colors,
            )

        if sensor_fn is not None:
            segment_program = self._make_gl_program(
                vertex_shader=_LINE_VERTEX_SHADER,
                geometry_shader=_LINE_GEOMETRY_SHADER,
                fragment_shader=_LINE_FRAGMENT_SHADER,
                color=(
                    np.array([0.0, 0.0, 0.0, 0.1], dtype=np.float32)
                    if sensor_color is None
                    else sensor_color
                ),
                width=np.array([sensor_width], dtype=np.float32),
                w_rad=np.array([sensor_width / 4], dtype=np.float32),
                l_rad=np.array([sensor_width / 4], dtype=np.float32),
            )

            def collect_sensors(stated: StateDict) -> NDArray:
                sensors = np.concatenate(
                    sensor_fn(stated=stated),  # type: ignore
                    axis=1,
                )
                sensors = sensors.reshape(-1, 2).astype(jnp.float32)
                flag = np.repeat(
                    np.array(stated.circle.is_active),
                    sensors.shape[0] // stated.circle.batch_size(),
                )
                return np.where(
                    flag.reshape(-1, 1),
                    sensors,
                    NOWHERE,
                )

            self._sensors = SegmentVA(
                ctx=context,
                program=segment_program,
                segments=collect_sensors(stated),
            )
            self._collect_sensors = collect_sensors
        else:
            self._sensors, self._collect_sensors = None, None

        head_program = self._make_gl_program(
            vertex_shader=_LINE_VERTEX_SHADER,
            geometry_shader=_LINE_GEOMETRY_SHADER,
            fragment_shader=_LINE_FRAGMENT_SHADER,
            color=np.array([0.5, 0.0, 1.0, 1.0], dtype=np.float32),
            width=np.array([0.004], dtype=np.float32),
        )
        self._heads = SegmentVA(
            ctx=context,
            program=head_program,
            segments=_collect_heads(space.shaped.circle, stated.circle),
        )

    def _make_gl_program(
        self,
        vertex_shader: str,
        geometry_shader: str | None = None,
        fragment_shader: str | None = None,
        screen_idx: tuple[int, int] = (0, 0),
        game_x: tuple[float, float] | None = None,
        game_y: tuple[float, float] | None = None,
        **kwargs: NDArray,
    ) -> mgl.Program:
        self._context.enable(mgl.PROGRAM_POINT_SIZE | mgl.BLEND)
        prog = self._context.program(
            vertex_shader=vertex_shader,
            geometry_shader=geometry_shader,
            fragment_shader=fragment_shader,
        )
        proj = _make_projection_matrix(
            game_x=game_x or (0, self._x_range),
            game_y=game_y or (0, self._y_range),
            screen_x=self._screen_x[screen_idx[0]],
            screen_y=self._screen_y[screen_idx[1]],
        )
        prog["proj"].write(proj)  # type: ignore
        for key, value in kwargs.items():
            prog[key].write(value)  # type: ignore
        return prog

    @staticmethod
    def _get_colors(default_colors: NDArray, colors: NDArray | None) -> NDArray:
        if colors is None:
            return default_colors
        else:
            clen = colors.shape[0]
            if clen < default_colors.shape[0]:
                return np.concatenate(
                    (colors.astype(np.float32), default_colors[clen:]),
                    axis=0,
                )
            else:
                return colors.astype(np.float32)

    def render(
        self,
        stated: StateDict,
        circle_colors: NDArray | None = None,
        sc_colors: NDArray | None = None,
        point_offset: NDArray | None = None,
    ) -> None:
        if point_offset is None:
            po = np.array([[0.0, 0.0]], dtype=np.float32)
        else:
            po = point_offset.astype(np.float32).reshape(1, 2)
        circle_points, circle_scale, circle_colors_default = _collect_circles(
            self._space.shaped.circle,
            stated.circle,
            self._circle_scaling,
        )
        if self._circles is not None:
            circle_colors = self._get_colors(circle_colors_default, circle_colors)
            if self._circles.update(circle_points + po, circle_scale, circle_colors):
                self._circles.render()
        if self._static_circles is not None:
            sc_points, sc_scale, _ = _collect_circles(
                self._space.shaped.static_circle,
                stated.static_circle,
                self._circle_scaling,
            )
            sc_colors = self._get_colors(
                _get_sc_color(self._sc_color, stated.static_circle),
                sc_colors,
            )
            if self._static_circles.update(sc_points + po, sc_scale, sc_colors):
                self._static_circles.render()
        if self._triangles is not None:
            points, _ = _collect_triangles(
                self._space.shaped.static_triangle,
                stated.static_triangle,
            )
            if self._triangles.update(points + po):
                self._triangles.render()
        if self._sensors is not None and self._collect_sensors is not None:
            if self._sensors.update(self._collect_sensors(stated) + po):
                self._sensors.render()
        if self._heads.update(
            _collect_heads(self._space.shaped.circle, stated.circle) + po
        ):
            self._heads.render()
        if self._static_lines is not None:
            if self._static_lines.update(self._static_line_points + po):
                self._static_lines.render()


class MglVisualizer:
    """
    Visualizer class that follows the `emevo.Visualizer` protocol.
    Considered as a main interface to use this visualizer.
    """

    def __init__(
        self,
        x_range: float,
        y_range: float,
        space: Space,
        stated: StateDict,
        sc_color: NDArray | None = None,
        sensor_color: NDArray | None = None,
        figsize: tuple[float, float] | None = None,
        voffsets: tuple[int, ...] = (),
        hoffsets: tuple[int, ...] = (),
        vsync: bool = False,
        backend: str = "pyglet",
        sensor_fn: Callable[[StateDict], tuple[NDArray, NDArray]] | None = None,
        sensor_width: float = 0.001,
        title: str = "EmEvo CircleForaging",
    ) -> None:
        self.pix_fmt = "rgba"

        if figsize is None:
            figsize = x_range * 3.0, y_range * 3.0
        w, h = int(figsize[0]), int(figsize[1])
        self._figsize = w + int(sum(hoffsets)), h + int(sum(voffsets))

        self._window = _make_window(
            title=title,
            size=self._figsize,
            backend=backend,
            vsync=vsync,
        )
        self._renderer = MglRenderer(
            context=self._window.ctx,
            screen_width=w,
            screen_height=h,
            x_range=x_range,
            y_range=y_range,
            space=space,
            stated=stated,
            voffsets=voffsets,
            hoffsets=hoffsets,
            sc_color_opt=sc_color,
            sensor_color=sensor_color,
            sensor_width=sensor_width,
            sensor_fn=sensor_fn,
        )

    def close(self) -> None:
        self._window.close()

    def get_image(self) -> NDArray:
        output = np.frombuffer(
            self._window.fbo.read(components=4, dtype="f1"),
            dtype=np.uint8,
        )
        w, h = self._figsize
        return output.reshape(h, w, -1)[::-1]

    def render(self, state: StateDict, **kwargs) -> None:
        self._window.clear(1.0, 1.0, 1.0)
        self._window.use()
        self._renderer.render(stated=state, **kwargs)

    def show(self) -> None:
        self._window.swap_buffers()


class _EglHeadlessWindow(headless.Window):
    name = "egl-headless"

    def init_mgl_context(self) -> None:
        """Create an standalone context and framebuffer"""
        self._ctx = mgl.create_standalone_context(
            require=self.gl_version_code,
            backend="egl",  # type: ignore
        )
        self._fbo = self.ctx.framebuffer(
            color_attachments=self.ctx.texture(self.size, 4, samples=self._samples),
            depth_attachment=self.ctx.depth_texture(self.size, samples=self._samples),
        )
        self.use()


def _make_window(
    *,
    title: str,
    size: tuple[int, int],
    backend: str,
    **kwargs,
) -> mglw.BaseWindow:
    if backend == "headless":
        window_cls = _EglHeadlessWindow
    else:
        window_cls = mglw.get_window_cls(f"moderngl_window.context.{backend}.Window")
    window = window_cls(title=title, gl_version=(4, 1), size=size, **kwargs)
    mglw.activate_context(ctx=window.ctx)
    return window


def _make_projection_matrix(
    game_x: tuple[float, float] = (0.0, 1.0),
    game_y: tuple[float, float] = (0.0, 1.0),
    screen_x: tuple[float, float] = (-1.0, 1.0),
    screen_y: tuple[float, float] = (-1.0, 1.0),
) -> NDArray:
    screen_width = screen_x[1] - screen_x[0]
    screen_height = screen_y[1] - screen_y[0]
    x_scale = screen_width / (game_x[1] - game_x[0])
    y_scale = screen_height / (game_y[1] - game_y[0])
    scale_mat = np.array(
        [
            [x_scale, 0, 0, 0],
            [0, y_scale, 0, 0],
            [0, 0, 0, 0],
            [0, 0, 0, 1],
        ],
        dtype=np.float32,
    )
    trans_mat = np.array(
        [
            [1, 0, 0, (sum(screen_x) - sum(game_x)) / screen_width],
            [0, 1, 0, (sum(screen_y) - sum(game_y)) / screen_height],
            [0, 0, 1, 0],
            [0, 0, 0, 1],
        ],
        dtype=np.float32,
    )
    return np.ascontiguousarray(np.dot(scale_mat, trans_mat).T)



import chex
import jax
import jax.numpy as jnp
import pytest

from phyjax2d import (
    Space,
    SpaceBuilder,
    Vec2d,
    circle_raycast,
    segment_raycast,
    thin_polygon_raycast,
)


@pytest.fixture
def space() -> Space:
    builder = SpaceBuilder(gravity=(0.0, 0.0))

    builder.add_circle(radius=2.0)
    builder.add_segment(p1=Vec2d(-2.0, -2.0), p2=Vec2d(2.0, 2.0))
    builder.add_polygon(points=[Vec2d(2.0, 0.0), Vec2d(0.0, 2.0), Vec2d(0.0, 0.0)])
    return builder.build()


def test_circle_raycast(space: Space) -> None:
    sd = space.zeros_state().nested_replace(
        "circle.p.xy",
        jnp.array([[0.0, 4.0]]),
    )
    p1 = jnp.array([[-4.0, 4.0], [-4.0, 3.0], [-4.0, 4.0]])
    p2 = jnp.array([[0.0, 4.0], [0.0, 3.0], [-8.0, 4.0]])
    rc = jax.vmap(circle_raycast, in_axes=(None, None, 0, 0, None, None))(
        0.0, 0.01, p1, p2, space.shaped.circle, sd.circle
    )
    chex.assert_trees_all_close(rc.fraction.ravel(), jnp.array([0.5, 0.5669873, -1.5]))


def test_segment_raycast(space: Space) -> None:
    sd = space.zeros_state()
    p1 = jnp.array([[-1.0, 1.0], [0.0, 2.0]])
    p2 = jnp.array([[1.0, -1.0], [1.0, 1.0]])
    rc = jax.vmap(segment_raycast, in_axes=(None, 0, 0, None, None))(
        0.01, p1, p2, space.shaped.segment, sd.segment
    )
    chex.assert_trees_all_close(rc.fraction.ravel(), jnp.array([0.5, 1.0]))


def test_polygon_raycast(space: Space) -> None:
    sd = space.zeros_state()
    p1 = jnp.array([[-2.0, 1.0], [0.0, -1.0]])
    p2 = jnp.array([[2.0, 1.0], [0.0, 9.0]])
    rc = jax.vmap(thin_polygon_raycast, in_axes=(None, 0, 0, None, None))(
        0.01, p1, p2, space.shaped.triangle, sd.triangle
    )
    chex.assert_trees_all_close(rc.fraction.ravel(), jnp.array([0.5, 0.1]))



import chex
import jax.numpy as jnp
import pytest

from phyjax2d import Space, SpaceBuilder, Vec2d, step


@pytest.fixture
def space() -> Space:
    builder = SpaceBuilder(gravity=(0.0, 0.0))

    builder.add_circle(radius=1.0)
    builder.add_circle(radius=2.0)
    builder.add_circle(radius=2.0)
    builder.add_segment(p1=Vec2d(-2.0, 0.0), p2=Vec2d(2.0, 0.0))
    builder.add_segment(p1=Vec2d(0.0, -2.0), p2=Vec2d(0.0, 2.0))
    builder.add_polygon(points=[Vec2d(2.0, 0.0), Vec2d(0.0, 2.0), Vec2d(0.0, 0.0)])
    builder.add_polygon(points=[Vec2d(2.0, 0.0), Vec2d(2.0, 2.0), Vec2d(0.0, 0.0)])
    return builder.build()


def test_circle_to_circle(space: Space) -> None:
    sd = space.zeros_state().nested_replace(
        "circle.p.xy",
        jnp.array([[1.0, 2.0], [3.0, 2.0], [4.0, 8.0]]),
    )
    contact = space.check_contacts_selected(sd, ("circle", "circle"))
    has_contact = contact.penetration >= 0
    chex.assert_trees_all_close(has_contact, jnp.array([True, False, False]))


def test_circle_to_segment(space: Space) -> None:
    sd = space.zeros_state().nested_replace(
        "circle.p.xy",
        jnp.array([[1.0, 2.0], [3.0, 2.0], [4.0, 8.0]]),
    )
    sd = sd.nested_replace("segment.p.xy", jnp.array([[0.0, 2.0], [4.0, 4.0]]))
    contact = space.check_contacts_selected(sd, ("segment", "circle"))
    has_contact = contact.penetration >= 0
    chex.assert_trees_all_close(
        has_contact,
        jnp.array([True, True, False, False, True, True]),
    )
    chex.assert_trees_all_close(
        contact.pos[:2],
        jnp.array([[1.0, 2.0], [1.5, 2.0]]),
    )
    chex.assert_trees_all_close(
        contact.pos[-2:],
        jnp.array([[4.5, 2.0], [4.0, 6.0]]),
    )


def test_circle_to_triangle(space: Space) -> None:
    sd = space.zeros_state().nested_replace(
        "circle.p.xy",
        jnp.array([[1.0, 2.0], [3.0, 2.0], [4.0, 5.9]]),
    )
    sd = sd.nested_replace("triangle.p.xy", jnp.array([[0.0, 3.5], [4.0, 4.0]]))
    contact = space.check_contacts_selected(sd, ("triangle", "circle"))
    has_contact = contact.penetration >= 0
    chex.assert_trees_all_close(
        has_contact,
        jnp.array([False, True, False, False, False, True]),
    )


def test_ignore() -> None:
    builder = SpaceBuilder(gravity=(0.0, 0.0))

    builder.add_circle(radius=1.0)
    builder.add_circle(radius=2.0, ignore=["circle"])
    builder.add_circle(radius=2.0, ignore=["static_circle"])
    builder.add_circle(radius=3.0, is_static=True)
    builder.add_circle(radius=3.0, is_static=True)
    space = builder.build()
    sd = space.zeros_state().nested_replace(
        "circle.p.xy",
        jnp.array([[1.0, 2.0], [3.0, 2.0], [4.0, 8.0]]),
    )
    sd = sd.nested_replace("static_circle.p.xy", jnp.array([[0.0, 2.0], [8.0, 8.0]]))
    contact = space.check_contacts_selected(sd, ("circle", "circle"))
    has_contact = contact.penetration >= 0
    chex.assert_trees_all_close(has_contact, jnp.array([False]))
    contact = space.check_contacts_selected(sd, ("circle", "static_circle"))
    has_contact = contact.penetration >= 0
    chex.assert_trees_all_close(has_contact, jnp.array([True, False, True, False]))

    solver = space.init_solver()
    step(space, sd, solver)



import jax.numpy as jnp

from phyjax2d import Position


def test_split() -> None:
    pos = Position(angle=jnp.zeros((10, 1)), xy=jnp.ones((10, 2)))
    p1, p2 = pos.split(3)
    assert len(p1.angle) == 3
    assert len(p1.xy) == 3
    assert len(p2.angle) == 7
    assert len(p2.xy) == 7



import chex
import jax.numpy as jnp
import pymunk
import pytest

from phyjax2d import Space, SpaceBuilder, Vec2d


@pytest.fixture
def space() -> Space:
    builder = SpaceBuilder(gravity=(0.0, 0.0))

    builder.add_circle(radius=2.0)
    builder.add_circle(radius=2.0)
    builder.add_polygon(points=[Vec2d(2.0, 0.0), Vec2d(0.0, 2.0), Vec2d(0.0, 0.0)])
    builder.add_polygon(
        points=[Vec2d(-1.0, 0.0), Vec2d(-3.0, 0.0), Vec2d(-2.0, -2.0)],
        radius=0.1,
    )
    return builder.build()


def test_triangle_to_circle(space: Space) -> None:
    sd = space.zeros_state().nested_replace(
        "circle.p.xy",
        jnp.array([[2.4, 2.4], [-2.0, -4.05]]),
    )
    contact = space.check_contacts_selected(sd, ("triangle", "circle"))
    has_contact = contact.penetration >= 0
    chex.assert_trees_all_close(has_contact, jnp.array([True, False, False, True]))
    sd = sd.nested_replace(
        "circle.p.xy",
        jnp.array([[2.42, 2.42], [-2.0, -4.11]]),
    )
    contact = space.check_contacts_selected(sd, ("triangle", "circle"))
    has_contact = contact.penetration >= 0
    chex.assert_trees_all_close(has_contact, jnp.array([False, False, False, False]))


def test_add_triangle() -> None:
    builder = SpaceBuilder(gravity=(0.0, 0.0))
    points = [Vec2d(2.0, 0.0), Vec2d(0.0, 2.0), Vec2d(-2.0, 0.0)]
    builder.add_polygon(points=points)
    polygon = builder.polygons[3][0]
    chex.assert_trees_all_close(
        polygon.points,
        jnp.array([[2.0, 0.0], [0.0, 2.0], [-2.0, 0.0]]),
    )
    chex.assert_trees_all_close(
        polygon.centroid,
        jnp.array([0.0, 2.0 / 3]),
    )
    root2 = 2**0.5
    chex.assert_trees_all_close(
        polygon.normals,
        jnp.array(
            [[1.0 / root2, 1.0 / root2], [-1.0 / root2, 1.0 / root2], [0.0, -1.0]]
        ),
    )
    chex.assert_trees_all_close(
        polygon.mass,
        jnp.array([4.0]),
    )
    chex.assert_trees_all_close(
        polygon.moment,
        jnp.array([pymunk.moment_for_poly(4.0, points)]),
    )


def test_add_square() -> None:
    builder = SpaceBuilder(gravity=(0.0, 0.0))
    builder.add_square(width=3.0, height=4.0)
    polygon = builder.polygons[4][0]
    chex.assert_trees_all_close(
        polygon.points,
        jnp.array([[1.5, 2.0], [-1.5, 2.0], [-1.5, -2.0], [1.5, -2.0]]),
    )
    chex.assert_trees_all_close(
        polygon.centroid,
        jnp.array([0.0, 0.0]),
    )
    chex.assert_trees_all_close(
        polygon.normals,
        jnp.array([[0.0, 1.0], [-1.0, 0.0], [0.0, -1.0], [1.0, 0.0]]),
    )
    chex.assert_trees_all_close(
        polygon.mass,
        jnp.array([12.0]),
    )
    chex.assert_trees_all_close(
        polygon.moment,
        jnp.array([pymunk.moment_for_box(12.0, (3.0, 4.0))]),
    )


def test_add_hexagon() -> None:
    builder = SpaceBuilder(gravity=(0.0, 0.0))
    points = [
        Vec2d(4.0, 2.0),
        Vec2d(2.0, 4.0),
        Vec2d(-2.0, 4.0),
        Vec2d(-4.0, -2.0),
        Vec2d(-2.0, -4.0),
        Vec2d(2.0, -4.0),
    ]
    builder.add_polygon(points=points)
    polygon = builder.polygons[6][0]
    chex.assert_trees_all_close(
        polygon.centroid,
        jnp.array([0.0, 0.0]),
    )
    chex.assert_trees_all_close(
        polygon.mass,
        jnp.array([48.0]),
    )
    chex.assert_trees_all_close(
        polygon.moment,
        jnp.array([pymunk.moment_for_poly(48.0, points)]),
    )



import jax
import jax.numpy as jnp
import typer

from phyjax2d import (
    Color,
    MglVisualizer,
    Space,
    SpaceBuilder,
    StateDict,
    Vec2d,
    VelocitySolver,
    make_square_segments,
)
from phyjax2d import step as phys_step

W, H = 600, 600

RED = Color(174, 68, 90)
PINK = Color(232, 188, 185)
NAVY = Color(75, 67, 118)
DARK_NAVY = Color(67, 46, 84)


def make_space() -> Space:
    builder = SpaceBuilder(gravity=(0.0, -9.8))
    walls = make_square_segments(0.0, 600.0, 0.0, 600.0)  # , rounded_offset=10.0
    builder.add_chain_segments(chain_points=walls)
    builder.add_circle(radius=40.0, color=RED)
    builder.add_circle(radius=40.0, color=RED)
    builder.add_circle(radius=60.0, color=PINK)
    builder.add_polygon(
        points=[Vec2d(40.0, 0.0), Vec2d(0.0, 40.0), Vec2d(0.0, 0.0)],
        color=NAVY,
    )
    builder.add_polygon(
        points=[Vec2d(30.0, 0.0), Vec2d(0.0, 30.0), Vec2d(-30.0, 0.0)],
        radius=0.1,
        color=DARK_NAVY,
    )
    return builder.build()


def main(n_steps: int = 100, scale: float = 1.0) -> None:
    space = make_space()
    sd = space.zeros_state().nested_replace(
        "circle.p.xy",
        jnp.array([[200.0, 300.0], [400.0, 300.0], [300.0, 400.0]]),
    )
    sd = sd.nested_replace(
        "triangle.p.xy",
        jnp.array([[150.0, 200.0], [500.0, 300.0]]),
    )
    sd = sd.nested_replace("triangle.p.angle", jnp.array([2.0, 1.0]))
    vis = MglVisualizer(
        x_range=W,
        y_range=H,
        space=space,
        stated=sd,
        figsize=(W * scale, H * scale),
    )
    vis.render(sd)
    vis.show()

    @jax.jit
    def step(sd: StateDict, solver: VelocitySolver) -> tuple[StateDict, VelocitySolver]:
        sd, solver, _ = phys_step(space, sd, solver)
        return sd, solver

    solver = space.init_solver()
    for _ in range(n_steps):
        sd, solver = step(sd, solver)
        vis.render(sd)
        vis.show()

    vis.close()


if __name__ == "__main__":
    app = typer.Typer(pretty_exceptions_show_locals=False)
    app.command()(main)
    app()



"""Physics implementation"""

from __future__ import annotations

import dataclasses
import functools
import uuid
from collections.abc import Sequence
from dataclasses import InitVar, replace
from typing import Any, Callable, Generic, Protocol, TypeVar

import chex
import jax
import jax.numpy as jnp

Self = Any
T = TypeVar("T")
TWO_PI = jnp.pi * 2


def then(x: Any, f: Callable[[Any], Any]) -> Any:
    if x is None:
        return x
    else:
        return f(x)


def normalize(
    x: jax.Array,
    axis: Sequence[int] | int | None = None,
) -> tuple[jax.Array, jax.Array]:
    norm = jnp.linalg.norm(x, axis=axis, keepdims=True)
    n = x / jnp.clip(norm, min=1e-6)
    return n, jnp.squeeze(norm, axis=axis)


def _vmap_dot(xy1: jax.Array, xy2: jax.Array) -> jax.Array:
    """Dot product between nested vectors"""
    chex.assert_equal_shape((xy1, xy2))
    orig_shape = xy1.shape
    a = xy1.reshape(-1, orig_shape[-1])
    b = xy2.reshape(-1, orig_shape[-1])
    return jax.vmap(jnp.dot, in_axes=(0, 0))(a, b).reshape(*orig_shape[:-1])


def _sv_cross(s: jax.Array, v: jax.Array) -> jax.Array:
    """Cross product with scalar and vector"""
    x, y = _get_xy(v)
    return jnp.stack((y * -s, x * s), axis=-1)


def empty(cls: type[T]) -> Callable[[], T]:
    all_fields = {f.name: jnp.empty(0) for f in dataclasses.fields(cls)}  # type: ignore
    return lambda: cls(**all_fields)


class PyTreeOps:
    def __add__(self, o: Any) -> Self:
        if o.__class__ is self.__class__:
            return jax.tree_util.tree_map(lambda x, y: x + y, self, o)
        else:
            return jax.tree_util.tree_map(lambda x: x + o, self)

    def __sub__(self, o: Any) -> Self:
        if o.__class__ is self.__class__:
            return jax.tree_util.tree_map(lambda x, y: x - y, self, o)
        else:
            return jax.tree_util.tree_map(lambda x: x - o, self)

    def __mul__(self, o: float | jax.Array) -> Self:
        return jax.tree_util.tree_map(lambda x: x * o, self)

    def __neg__(self) -> Self:
        return jax.tree_util.tree_map(lambda x: -x, self)

    def __truediv__(self, o: float | jax.Array) -> Self:
        return jax.tree_util.tree_map(lambda x: x / o, self)

    @jax.jit
    def get_slice(self, index: Sequence[int] | int | None) -> Self:
        return jax.tree_util.tree_map(lambda x: x[index], self)

    def split(self, split_index: int) -> tuple[Self, Self]:
        def split_inner(x: jax.Array) -> tuple[jax.Array, jax.Array]:
            return x[:split_index], x[split_index:]

        structure = jax.tree.structure(self)
        return jax.tree.transpose(
            structure,
            jax.tree.structure(("*", "*")),
            jax.tree.map(split_inner, self),
        )

    def reshape(self, shape: Sequence[int]) -> Self:
        return jax.tree_util.tree_map(lambda x: x.reshape(shape), self)

    def sum(self, axis: int | None = None) -> Self:
        return jax.tree_util.tree_map(lambda x: jnp.sum(x, axis=axis), self)

    def tolist(self) -> list[Self]:
        leaves, treedef = jax.tree_util.tree_flatten(self)
        return [treedef.unflatten(leaf) for leaf in zip(*leaves)]

    def zeros_like(self) -> Any:
        return jax.tree_util.tree_map(lambda x: jnp.zeros_like(x), self)

    @property
    def shape(self) -> Any:
        """For debugging"""
        return jax.tree_util.tree_map(lambda x: x.shape, self)


def _axy(angle: jax.Array, xy: jax.Array) -> jax.Array:
    return jnp.concatenate((jnp.expand_dims(angle, axis=-1), xy), axis=-1)


class _PositionLike(Protocol):
    angle: jax.Array  # Angular velocity (N,)
    xy: jax.Array  # (N, 2)

    def __init__(self, angle: jax.Array, xy: jax.Array) -> None: ...

    def batch_size(self) -> int:
        return self.angle.shape[0]

    def into_axy(self) -> jax.Array:
        return _axy(self.angle, self.xy)

    @classmethod
    def zeros(cls: type[Self], n: int) -> Self:
        return cls(angle=jnp.zeros((n,)), xy=jnp.zeros((n, 2)))

    @classmethod
    def from_axy(cls: type[Self], axy: jax.Array) -> Self:
        angle = jax.lax.squeeze(jax.lax.slice_in_dim(axy, 0, 1, axis=-1), (-1,))
        xy = jax.lax.slice_in_dim(axy, 1, 3, axis=-1)
        return cls(angle=angle, xy=xy)


@chex.dataclass
class Velocity(_PositionLike, PyTreeOps):
    angle: jax.Array  # Angular velocity (N,)
    xy: jax.Array  # (N, 2)

    def rv(self, r: jax.Array) -> jax.Array:
        """Relative velocity"""
        return self.xy + _sv_cross(self.angle, r)


@chex.dataclass
class Force(_PositionLike, PyTreeOps):
    angle: jax.Array  # Angular (torque) force (N,)
    xy: jax.Array  # (N, 2)


def _get_xy(xy: jax.Array) -> tuple[jax.Array, jax.Array]:
    x = jax.lax.slice_in_dim(xy, 0, 1, axis=-1)
    y = jax.lax.slice_in_dim(xy, 1, 2, axis=-1)
    return jax.lax.squeeze(x, (-1,)), jax.lax.squeeze(y, (-1,))


def _right_perp(xy: jax.Array) -> jax.Array:
    x = jax.lax.slice_in_dim(xy, 0, 1, axis=-1)
    y = jax.lax.slice_in_dim(xy, 1, 2, axis=-1)
    return jnp.concatenate((y, -x))


@chex.dataclass
class Position(_PositionLike, PyTreeOps):
    angle: jax.Array  # Angular velocity (N, 1)
    xy: jax.Array  # (N, 2)

    def rotate(self, xy: jax.Array) -> jax.Array:
        x, y = _get_xy(xy)
        s, c = jnp.sin(self.angle), jnp.cos(self.angle)
        rot_x = c * x - s * y
        rot_y = s * x + c * y
        return jnp.stack((rot_x, rot_y), axis=-1)

    def transform(self, xy: jax.Array) -> jax.Array:
        return self.rotate(xy) + self.xy

    def inv_rotate(self, xy: jax.Array) -> jax.Array:
        x, y = _get_xy(xy)
        s, c = jnp.sin(self.angle), jnp.cos(self.angle)
        rot_x = c * x + s * y
        rot_y = c * y - s * x
        return jnp.stack((rot_x, rot_y), axis=-1)

    def inv_transform(self, xy: jax.Array) -> jax.Array:
        return self.inv_rotate(xy - self.xy)

    def to_origin(self) -> Self:
        return replace(self, xy=jnp.zeros_like(self.xy))


@chex.dataclass
class Shape(PyTreeOps):
    mass: jax.Array
    moment: jax.Array
    elasticity: jax.Array
    friction: jax.Array
    rgba: jax.Array

    def batch_size(self) -> int:
        return self.mass.shape[0]

    def is_empty(self) -> bool:
        return self.batch_size() == 0

    def inv_mass(self) -> jax.Array:
        """To support static shape, set let inv_mass 0 if mass is infinite"""
        return jnp.where(jnp.isfinite(self.mass), 1.0 / self.mass, 0.0)

    def inv_moment(self) -> jax.Array:
        """As inv_mass does, set inv_moment 0 if moment is infinite"""
        return jnp.where(jnp.isfinite(self.moment), 1.0 / self.moment, 0.0)

    def to_shape(self) -> Self:
        return Shape(
            mass=self.mass,
            moment=self.moment,
            elasticity=self.elasticity,
            friction=self.friction,
            rgba=self.rgba,
        )


@chex.dataclass
class Circle(Shape):
    radius: jax.Array


@chex.dataclass
class Capsule(Shape):
    point1: jax.Array
    point2: jax.Array
    radius: jax.Array


@chex.dataclass
class Segment(Shape):
    point1: jax.Array
    point2: jax.Array
    is_smooth: jax.Array
    ghost1: jax.Array
    ghost2: jax.Array


@chex.dataclass
class Polygon(Shape):
    points: jax.Array
    radius: jax.Array
    normals: jax.Array
    centroid: jax.Array


@chex.dataclass
class Contact(PyTreeOps):
    pos: jax.Array
    normal: jax.Array
    penetration: jax.Array
    elasticity: jax.Array
    friction: jax.Array

    def contact_dim(self) -> int:
        return self.pos.shape[1]


@jax.vmap
def _circle_to_circle_impl(
    a: Circle,
    b: Circle,
    a_pos: Position,
    b_pos: Position,
    isactive: jax.Array,
) -> Contact:
    a2b_normal, dist = normalize(b_pos.xy - a_pos.xy)
    penetration = a.radius + b.radius - dist
    a_contact = a_pos.xy + a2b_normal * a.radius
    b_contact = b_pos.xy - a2b_normal * b.radius
    pos = (a_contact + b_contact) * 0.5
    penetration = jnp.where(isactive, penetration, -1.0)
    return Contact(
        pos=pos,
        normal=a2b_normal,
        penetration=penetration,
        elasticity=(a.elasticity + b.elasticity) * 0.5,
        friction=(a.friction + b.friction) * 0.5,
    )


@chex.dataclass
class ContactHelper:
    tangent: jax.Array
    mass_normal: jax.Array
    mass_tangent: jax.Array
    v_bias: jax.Array
    bounce: jax.Array
    r1: jax.Array
    r2: jax.Array
    inv_mass1: jax.Array
    inv_mass2: jax.Array
    inv_moment1: jax.Array
    inv_moment2: jax.Array
    local_anchor1: jax.Array
    local_anchor2: jax.Array
    allow_bounce: jax.Array


def _effective_mass(
    inv_mass: jax.Array,
    inv_moment: jax.Array,
    r: jax.Array,
    n: jax.Array,
) -> jax.Array:
    rn2 = jnp.cross(r, n) ** 2
    return inv_mass + inv_moment * rn2


@jax.vmap
def _capsule_to_circle_impl(
    a: Capsule,
    b: Circle,
    a_pos: Position,
    b_pos: Position,
    isactive: jax.Array,
) -> Contact:
    # Move b_pos to capsule's coordinates
    pb = a_pos.inv_transform(b_pos.xy)
    p1, p2 = a.point1, a.point2
    edge = p2 - p1
    s1 = jnp.dot(pb - p1, edge)
    s2 = jnp.dot(p2 - pb, edge)
    in_segment = jnp.logical_and(s1 >= 0.0, s2 >= 0.0)
    ee = jnp.sum(jnp.square(edge), axis=-1, keepdims=True)
    # Closest point
    # s1 < 0: pb is left to the capsule
    # s2 < 0: pb is right to the capsule
    # else: pb is in between capsule
    pa = jax.lax.select(
        in_segment,
        p1 + edge * s1 / ee,
        jax.lax.select(s1 < 0.0, p1, p2),
    )
    a2b_normal, dist = normalize(pb - pa)
    penetration = a.radius + b.radius - dist
    a_contact = pa + a2b_normal * a.radius
    b_contact = pb - a2b_normal * b.radius
    pos = a_pos.transform((a_contact + b_contact) * 0.5)
    a2b_normal_rotated = a_pos.to_origin().transform(a2b_normal)
    # Filter penetration
    return Contact(
        pos=pos,
        normal=a2b_normal_rotated,
        penetration=jnp.where(isactive, penetration, -1.0),
        elasticity=(a.elasticity + b.elasticity) * 0.5,
        friction=(a.friction + b.friction) * 0.5,
    )


@jax.vmap
def _segment_to_circle_impl(
    a: Segment,
    b: Circle,
    a_pos: Position,
    b_pos: Position,
    isactive: jax.Array,
) -> Contact:
    # Move b_pos to segment's coordinates
    pb = a_pos.inv_transform(b_pos.xy)
    p1, p2 = a.point1, a.point2
    edge = p2 - p1
    s1 = jnp.dot(pb - p1, edge)
    s2 = jnp.dot(p2 - pb, edge)
    in_segment = jnp.logical_and(s1 > 0.0, s2 > 0.0)
    ee = jnp.sum(jnp.square(edge), axis=-1, keepdims=True)
    # Closest point
    # s1 < 0: pb is left to the capsule
    # s2 < 0: pb is right to the capsule
    # else: pb is in between capsule
    pa = jax.lax.select(
        in_segment,
        p1 + edge * s1 / ee,
        jax.lax.select(s1 <= 0.0, p1, p2),
    )
    a2b_normal, dist = normalize(pb - pa)
    penetration = b.radius - dist
    a_contact = pa
    b_contact = pb - a2b_normal * b.radius
    pos = a_pos.transform((a_contact + b_contact) * 0.5)
    # Filter penetration
    collidable = jnp.dot(_right_perp(edge), pb - p1) >= 0.0
    not_in_voronoi = jnp.logical_or(
        jnp.logical_and(s1 <= 0.0, jnp.dot(a.ghost2 - p2, pb - p2) > 0.0),
        jnp.logical_and(s2 <= 0.0, jnp.dot(p1 - a.ghost1, pb - p1) <= 0.0),
    )
    is_penetration_possible = jnp.logical_and(
        isactive,
        jnp.logical_or(
            jnp.logical_not(a.is_smooth),
            # collidable
            jnp.logical_and(collidable, jnp.logical_not(not_in_voronoi)),
        ),
    )
    return Contact(
        pos=pos,
        normal=a_pos.to_origin().transform(a2b_normal),
        penetration=jnp.where(is_penetration_possible, penetration, -1.0),
        elasticity=(a.elasticity + b.elasticity) * 0.5,
        friction=(a.friction + b.friction) * 0.5,
    )


@jax.vmap
def _polygon_to_circle_impl(
    a: Polygon,
    b: Circle,
    a_pos: Position,
    b_pos: Position,
    isactive: jax.Array,
) -> Contact:
    n_vertices = a.points.shape[0]
    # Move circle pos to segment's coordinates
    c = a_pos.inv_transform(b_pos.xy)  # (1, 2)
    separation = _vmap_dot(a.normals, (c - a.points))
    max_sep = jnp.max(separation)
    i1 = jnp.argmax(separation)
    i2 = (i1 + 1) % n_vertices
    v1 = a.points[i1]
    v2 = a.points[i2]
    u1 = jnp.squeeze(jnp.dot(c - v1, v2 - v1))  # Convert to scalar
    u2 = jnp.squeeze(jnp.dot(c - v2, v1 - v2))

    def if_c_is_out() -> Contact:
        v = jax.lax.select(u1 < 0.0, v1, v2)
        a2b_normal, dist = normalize(c - v)
        ca = v + a.radius * a2b_normal
        cb = c - b.radius * a2b_normal
        pos_a = (ca + cb) * 0.5
        penetration = a.radius + b.radius - dist
        return Contact(
            pos=pos_a + a_pos.xy,
            normal=a_pos.to_origin().transform(a2b_normal),
            penetration=jnp.where(isactive, penetration, -1.0),
            elasticity=(a.elasticity + b.elasticity) * 0.5,
            friction=(a.friction + b.friction) * 0.5,
        )

    def if_c_is_in() -> Contact:
        a2b_normal = a.normals[i1]
        ca = c + a.radius - jnp.dot(c - v1, a2b_normal) * a2b_normal
        cb = c - b.radius * a2b_normal
        pos_a = (ca + cb) * 0.5
        penetration = a.radius + b.radius - max_sep
        return Contact(
            pos=pos_a + a_pos.xy,
            normal=a_pos.to_origin().transform(a2b_normal),
            penetration=jnp.where(isactive, penetration, -1.0),
            elasticity=(a.elasticity + b.elasticity) * 0.5,
            friction=(a.friction + b.friction) * 0.5,
        )

    is_out = jnp.logical_or(u1 < 0.0, u2 < 0.0)
    return jax.lax.cond(
        jnp.logical_and(is_out, max_sep > 1e-6),
        if_c_is_out,
        if_c_is_in,
    )


_ALL_SHAPES = [
    "circle",
    "static_circle",
    "capsule",
    "static_capsule",
    "segment",
    "triangle",
    "static_triangle",
    "quadrangle",
    "static_quadrangle",
    "pentagon",
    "static_pentagon",
    "hexagon",
    "static_hexagon",
]


@chex.dataclass
class State(PyTreeOps):
    p: Position
    v: Velocity
    f: Force
    is_active: jax.Array
    label: jax.Array

    @staticmethod
    def empty() -> Self:
        return State(
            p=Position.zeros(0),
            v=Velocity.zeros(0),
            f=Force.zeros(0),
            is_active=jnp.empty(0),
            label=jnp.zeros(0),
        )

    @staticmethod
    def zeros(n: int) -> Self:
        return State(
            p=Position.zeros(n),
            v=Velocity.zeros(n),
            f=Force.zeros(n),
            is_active=jnp.ones(n, dtype=bool),
            label=jnp.zeros(n, dtype=jnp.uint8),
        )

    def is_empty(self) -> bool:
        return self.p.batch_size() == 0

    def apply_force_global(self, point: jax.Array, force: jax.Array) -> Self:
        chex.assert_equal_shape((self.f.xy, force))
        xy = self.f.xy + force
        angle = self.f.angle + jnp.cross(point - self.p.xy, force)
        f = replace(self.f, xy=xy, angle=angle)
        return replace(self, f=f)

    def apply_force_local(self, point: jax.Array, force: jax.Array) -> Self:
        chex.assert_equal_shape((self.p.xy, point))
        point = self.p.transform(point)
        return self.apply_force_global(point, self.p.rotate(force))

    def batch_size(self) -> int:
        return self.p.batch_size()


def get_relative_angle(s_a: State, s_b: State) -> jax.Array:
    a2b = jax.vmap(jnp.subtract, in_axes=(None, 0))(s_b.p.xy, s_a.p.xy)
    a2b_x, a2b_y = _get_xy(a2b)
    a2b_angle = jnp.arctan2(a2b_y, a2b_x)  # (N_A, N_B)
    a_angle = jnp.expand_dims(s_a.p.angle, axis=1)
    # Subtract 0.5𝛑 because our angle starts from 0.5𝛑 (90 degree)
    return (a2b_angle - a_angle + TWO_PI * 3 - jnp.pi * 0.5) % TWO_PI


def _offset(sd: ShapeDict | StateDict, name: str) -> int:
    total = 0
    for key in _ALL_SHAPES:
        if key == name:
            return total
        s = sd[key]  # type: ignore
        if s is not None:
            total += s.batch_size()
    raise RuntimeError("Unreachable")


@chex.dataclass
class StateDict:
    circle: State = dataclasses.field(default_factory=State.empty)
    static_circle: State = dataclasses.field(default_factory=State.empty)
    segment: State = dataclasses.field(default_factory=State.empty)
    capsule: State = dataclasses.field(default_factory=State.empty)
    static_capsule: State = dataclasses.field(default_factory=State.empty)
    triangle: State = dataclasses.field(default_factory=State.empty)
    static_triangle: State = dataclasses.field(default_factory=State.empty)
    quadrangle: State = dataclasses.field(default_factory=State.empty)
    static_quadrangle: State = dataclasses.field(default_factory=State.empty)
    pentagon: State = dataclasses.field(default_factory=State.empty)
    static_pentagon: State = dataclasses.field(default_factory=State.empty)
    hexagon: State = dataclasses.field(default_factory=State.empty)
    static_hexagon: State = dataclasses.field(default_factory=State.empty)

    def concat(self) -> Self:
        states = [s for s in self.values() if s.batch_size() > 0]  # type: ignore
        return jax.tree_util.tree_map(
            lambda *args: jnp.concatenate(args, axis=0), *states
        )

    def update(self, statec: State) -> Self:
        def _get(name: str) -> State:
            state = self[name]  # type: ignore
            if state.is_empty():
                return state  # empty state
            else:
                start = _offset(self, name)
                end = start + state.p.batch_size()
                return statec.get_slice(jnp.arange(start, end))

        sd = {key: _get(key) for key in _ALL_SHAPES}
        return self.__class__(**sd)

    def nested_replace(self, query: str, value: Any) -> Self:
        """Convenient method for nested replace"""
        queries = query.split(".")
        objects = [self]
        for q in queries[:-1]:
            objects.append(objects[-1][q])  # type: ignore
        obj = replace(objects[-1], **{queries[-1]: value})
        for o, q in zip(objects[-2::-1], queries[-2::-1]):
            obj = replace(o, **{q: obj})
        return obj

    def nested_ops(self, query: str, fn: Callable[[Any], Any]) -> Self:
        queries = query.split(".")
        objects = [self]
        for q in queries[:-1]:
            objects.append(objects[-1][q])  # type: ignore
        value = fn(objects[-1][queries[-1]])  # type: ignore
        obj = replace(objects[-1], **{queries[-1]: value})
        for o, q in zip(objects[-2::-1], queries[-2::-1]):
            obj = replace(o, **{q: obj})
        return obj


@chex.dataclass
class ShapeDict:
    circle: Circle = dataclasses.field(default_factory=empty(Circle))
    static_circle: Circle = dataclasses.field(default_factory=empty(Circle))
    capsule: Capsule = dataclasses.field(default_factory=empty(Capsule))
    static_capsule: Capsule = dataclasses.field(default_factory=empty(Capsule))
    segment: Segment = dataclasses.field(default_factory=empty(Segment))
    triangle: Polygon = dataclasses.field(default_factory=empty(Polygon))
    static_triangle: Polygon = dataclasses.field(default_factory=empty(Polygon))
    quadrangle: Polygon = dataclasses.field(default_factory=empty(Polygon))
    static_quadrangle: Polygon = dataclasses.field(default_factory=empty(Polygon))
    pentagon: Polygon = dataclasses.field(default_factory=empty(Polygon))
    static_pentagon: Polygon = dataclasses.field(default_factory=empty(Polygon))
    hexagon: Polygon = dataclasses.field(default_factory=empty(Polygon))
    static_hexagon: Polygon = dataclasses.field(default_factory=empty(Polygon))

    def concat(self) -> Shape:
        shapes = [
            s.to_shape()
            for s in self.values()
            if s.batch_size() > 0  # type: ignore
        ]
        return jax.tree_util.tree_map(
            lambda *args: jnp.concatenate(args, axis=0), *shapes
        )

    def n_shapes(self) -> int:
        return sum([s.batch_size() for s in self.values()])  # type: ignore

    def zeros_state(self) -> StateDict:
        circle = then(self.circle, lambda s: State.zeros(len(s.mass)))
        static_circle = then(self.static_circle, lambda s: State.zeros(len(s.mass)))
        segment = then(self.segment, lambda s: State.zeros(len(s.mass)))
        capsule = then(self.capsule, lambda s: State.zeros(len(s.mass)))
        static_capsule = then(self.capsule, lambda s: State.zeros(len(s.mass)))
        triangle = then(self.triangle, lambda s: State.zeros(len(s.mass)))
        static_triangle = then(self.static_triangle, lambda s: State.zeros(len(s.mass)))
        quadrangle = then(self.quadrangle, lambda s: State.zeros(len(s.mass)))
        static_quadrangle = then(
            self.static_quadrangle,
            lambda s: State.zeros(len(s.mass)),
        )
        pentagon = then(self.pentagon, lambda s: State.zeros(len(s.mass)))
        static_pentagon = then(self.static_pentagon, lambda s: State.zeros(len(s.mass)))
        hexagon = then(self.hexagon, lambda s: State.zeros(len(s.mass)))
        static_hexagon = then(self.static_hexagon, lambda s: State.zeros(len(s.mass)))
        return StateDict(
            circle=circle,
            static_circle=static_circle,
            segment=segment,
            capsule=capsule,
            static_capsule=static_capsule,
            triangle=triangle,
            static_triangle=static_triangle,
            quadrangle=quadrangle,
            static_quadrangle=static_quadrangle,
            pentagon=pentagon,
            static_pentagon=static_pentagon,
            hexagon=hexagon,
            static_hexagon=static_hexagon,
        )


S1 = TypeVar("S1", bound=Shape)
S2 = TypeVar("S2", bound=Shape)


@chex.dataclass
class ContactIndices(Generic[S1, S2]):
    shape1: S1
    shape2: S2
    index1: jax.Array
    index2: jax.Array


@dataclasses.dataclass
class _ContactConstraint:
    s1_ignore: jax.Array
    s2_ignore: jax.Array


def _self_ci(shape: Shape, constraint: _ContactConstraint) -> ContactIndices:
    n = shape.batch_size() - len(constraint.s1_ignore)
    index1, index2 = jax.jit(jnp.triu_indices, static_argnums=(0, 1))(n, 1)

    all_indices = jnp.arange(shape.batch_size())
    mapping = all_indices[~jnp.isin(all_indices, constraint.s1_ignore)]
    index1 = mapping[index1]
    index2 = mapping[index2]
    return ContactIndices(
        shape1=shape.get_slice(index1),
        shape2=shape.get_slice(index2),
        index1=index1,
        index2=index2,
    )


def _pair_ci(
    shape1: Shape,
    shape2: Shape,
    constraint: _ContactConstraint,
) -> ContactIndices:
    @functools.partial(jax.jit, static_argnums=(1,))
    def pair_outer(x: jax.Array, reps: int) -> jax.Array:
        return jnp.repeat(x, reps, axis=0, total_repeat_length=x.shape[0] * reps)

    @functools.partial(jax.jit, static_argnums=(1,))
    def pair_inner(x: jax.Array, reps: int) -> jax.Array:
        return jnp.tile(x, (reps,) + (1,) * (x.ndim - 1))

    def _remove_ignored_indices(
        mapped_array: jax.Array,
        ignore_array: jax.Array,
    ) -> jax.Array:
        n_ignore = len(ignore_array)
        if n_ignore == 0:
            return mapped_array
        all_indices = jnp.arange(len(mapped_array) + n_ignore)
        mapping = all_indices[~jnp.isin(all_indices, ignore_array)]
        return mapping[mapped_array]

    n1 = shape1.batch_size() - len(constraint.s1_ignore)
    n2 = shape2.batch_size() - len(constraint.s2_ignore)
    index1 = pair_outer(jnp.arange(n1), reps=n2)
    index2 = pair_inner(jnp.arange(n2), reps=n1)
    index1 = _remove_ignored_indices(index1, constraint.s1_ignore)
    index2 = _remove_ignored_indices(index2, constraint.s2_ignore)
    return ContactIndices(
        shape1=shape1.get_slice(index1),
        shape2=shape2.get_slice(index2),
        index1=index1,
        index2=index2,
    )


def _circle_to_circle(ci: ContactIndices[Circle, Circle], stated: StateDict) -> Contact:
    return _circle_to_circle_impl(
        ci.shape1,
        ci.shape2,
        jax.tree_util.tree_map(lambda arr: arr[ci.index1], stated.circle.p),
        jax.tree_util.tree_map(lambda arr: arr[ci.index2], stated.circle.p),
        jnp.logical_and(
            stated.circle.is_active[ci.index1],
            stated.circle.is_active[ci.index2],
        ),
    )


def _circle_to_static_circle(
    ci: ContactIndices[Circle, Circle],
    stated: StateDict,
) -> Contact:
    return _circle_to_circle_impl(
        ci.shape1,
        ci.shape2,
        jax.tree_util.tree_map(lambda arr: arr[ci.index1], stated.circle.p),
        jax.tree_util.tree_map(lambda arr: arr[ci.index2], stated.static_circle.p),
        jnp.logical_and(
            stated.circle.is_active[ci.index1],
            stated.static_circle.is_active[ci.index2],
        ),
    )


def _capsule_to_circle(
    ci: ContactIndices[Capsule, Circle],
    stated: StateDict,
) -> Contact:
    return _capsule_to_circle_impl(
        ci.shape1,
        ci.shape2,
        jax.tree_util.tree_map(lambda arr: arr[ci.index1], stated.capsule.p),
        jax.tree_util.tree_map(lambda arr: arr[ci.index2], stated.circle.p),
        jnp.logical_and(
            stated.capsule.is_active[ci.index1],
            stated.circle.is_active[ci.index2],
        ),
    )


def _segment_to_circle(
    ci: ContactIndices[Segment, Circle],
    stated: StateDict,
) -> Contact:
    return _segment_to_circle_impl(
        ci.shape1,
        ci.shape2,
        jax.tree_util.tree_map(lambda arr: arr[ci.index1], stated.segment.p),
        jax.tree_util.tree_map(lambda arr: arr[ci.index2], stated.circle.p),
        jnp.logical_and(
            stated.segment.is_active[ci.index1], stated.circle.is_active[ci.index2]
        ),
    )


def _polygon_to_circle(
    ci: ContactIndices[Polygon, Circle],
    stated: StateDict,
    *,
    key: str,
) -> Contact:
    polygon = stated[key]  # type: ignore
    return _polygon_to_circle_impl(
        ci.shape1,
        ci.shape2,
        jax.tree_util.tree_map(lambda arr: arr[ci.index1], polygon.p),
        jax.tree_util.tree_map(lambda arr: arr[ci.index2], stated.circle.p),
        jnp.logical_and(
            polygon.is_active[ci.index1],
            stated.circle.is_active[ci.index2],
        ),
    )


def _polygon_to_polygon(
    ci: ContactIndices[Polygon, Polygon],
    stated: StateDict,
    *,
    key1: str,
    key2: str,
) -> Contact:
    assert False


_CONTACT_FN = Callable[[ContactIndices, StateDict], Contact]
_CONTACT_FUNCTIONS: dict[tuple[str, str], _CONTACT_FN] = {
    ("circle", "circle"): _circle_to_circle,
    ("circle", "static_circle"): _circle_to_static_circle,
    ("capsule", "circle"): _capsule_to_circle,
    ("segment", "circle"): _segment_to_circle,
    ("triangle", "circle"): functools.partial(_polygon_to_circle, key="triangle"),
    ("static_triangle", "circle"): functools.partial(
        _polygon_to_circle,
        key="static_triangle",
    ),
    # ("triangle", "triangle"): functools.partial(
    #     _polygon_to_polygon,
    #     key1="triangle",
    #     key2="triangle",
    # ),
}


# TODO: pretty print
@chex.dataclass
class Space:
    gravity: jax.Array
    shaped: ShapeDict
    dt: float = 0.1
    linear_damping: float = 0.95
    angular_damping: float = 0.95
    bias_factor: float = 0.2
    n_velocity_iter: int = 6
    n_position_iter: int = 2
    linear_slop: float = 0.005
    speculative_distance: float = 0.02
    max_linear_correction: float = 0.2
    allowed_penetration: float = 0.005
    bounce_threshold: float = 1.0
    max_velocity: float = 100.0
    max_angular_velocity: float = 100.0
    ignore_collision: InitVar[dict[tuple[str, str], list[int]] | None] = None
    _contact_offset: dict[tuple[str, str], tuple[int, int]] = dataclasses.field(
        default_factory=dict,
        init=False,
    )
    _n_contacts: dict[tuple[str, str], int] = dataclasses.field(
        default_factory=dict,
        init=False,
    )
    _ci: dict[tuple[str, str], ContactIndices] = dataclasses.field(
        default_factory=dict,
        init=False,
    )
    _cc: dict[tuple[str, str], _ContactConstraint] = dataclasses.field(
        default_factory=dict,
        init=False,
    )
    _ci_total: ContactIndices = dataclasses.field(init=False)
    _hash_key: uuid.UUID = dataclasses.field(default_factory=uuid.uuid4, init=False)

    def __hash__(self) -> int:
        return hash(self._hash_key)

    def __eq__(self, other: Any) -> bool:
        return self._hash_key == other._hash_key

    def __post_init__(
        self,
        ignore_collision: dict[tuple[str, str], list] | None,
    ) -> None:
        def get_collision_constraint(n1: str, n2: str) -> _ContactConstraint:
            if ignore_collision is None:
                return _ContactConstraint(
                    s1_ignore=jnp.array([]),
                    s2_ignore=jnp.array([]),
                )
            ignore_n1 = ignore_collision.get((n1, n2), [])
            ignore_n2 = ignore_collision.get((n2, n1), [])
            return _ContactConstraint(jnp.array(ignore_n1), jnp.array(ignore_n2))

        ci_slided_list = []
        offset = 0
        total_ci_len = 0
        for n1, n2 in _CONTACT_FUNCTIONS.keys():
            shape1, shape2 = self.shaped[n1], self.shaped[n2]  # type: ignore
            if shape1.batch_size() > 0 and shape2.batch_size() > 0:
                cc = get_collision_constraint(n1, n2)
                if n1 == n2:
                    ci = _self_ci(shape1, constraint=cc)  # Type
                else:
                    ci = _pair_ci(shape1, shape2, constraint=cc)
                self._cc[n1, n2] = cc
                self._ci[n1, n2] = ci
                n_contacts = ci.index1.shape[0]
                total_ci_len += n_contacts
                self._n_contacts[n1, n2] = n_contacts
                offset_start = offset
                offset += n_contacts
                self._contact_offset[n1, n2] = offset_start, offset
                offset1, offset2 = _offset(self.shaped, n1), _offset(self.shaped, n2)
                # Add some offset for global indices
                ci_slided = ContactIndices(
                    shape1=ci.shape1.to_shape(),
                    shape2=ci.shape2.to_shape(),
                    index1=ci.index1 + offset1,
                    index2=ci.index2 + offset2,
                )
                ci_slided_list.append(ci_slided)

        self._ci_total = jax.tree.map(
            lambda *args: jnp.concatenate(args, axis=0),
            *ci_slided_list,
        )

    def check_contacts(self, stated: StateDict) -> Contact:
        contacts = []
        for (n1, n2), fn in _CONTACT_FUNCTIONS.items():
            ci = self._ci.get((n1, n2), None)
            if ci is not None:
                contact = fn(ci, stated)
                contacts.append(contact)
        return jax.tree.map(
            lambda *args: jnp.concatenate(args, axis=0),
            *contacts,
        )

    def check_contacts_selected(
        self,
        stated: StateDict,
        selected: tuple[str, str],
    ) -> Contact | None:
        fn = _CONTACT_FUNCTIONS[selected]
        ci = self._ci.get(selected, None)
        if ci is not None:
            return fn(ci, stated)
        else:
            return None

    def get_contact_mat(self, n1: str, n2: str, contact: jax.Array) -> jax.Array:
        contact_offset = self._contact_offset.get((n1, n2), None)
        assert contact_offset is not None
        from_, to = contact_offset
        cc = self._cc[n1, n2]
        size1 = self.shaped[n1].batch_size() - len(cc.s1_ignore)  # type: ignore
        size2 = self.shaped[n2].batch_size() - len(cc.s2_ignore)  # type: ignore
        cnt = contact[from_:to]
        if n1 == n2:
            ret = jnp.zeros((size1, size1), dtype=bool)
            idx1, idx2 = jnp.triu_indices(size1, k=1)
            return ret.at[idx1, idx2].set(cnt).at[idx2, idx1].set(cnt)
        else:
            return contact[from_:to].reshape(size1, size2)

    def init_solver(self) -> VelocitySolver:
        n = sum(self._n_contacts.values())
        return VelocitySolver(
            v1=jnp.zeros((n, 3)),
            v2=jnp.zeros((n, 3)),
            pn=jnp.zeros(n),
            pt=jnp.zeros(n),
            contact=jnp.zeros(n, dtype=bool),
        )

    def zeros_state(self) -> StateDict:
        return self.shaped.zeros_state()


def update_velocity(space: Space, shape: Shape, state: State) -> State:
    # Expand (N, ) to (N, 1) because xy has a shape (N, 2)
    invm = jnp.expand_dims(shape.inv_mass(), axis=1)
    f_xy = jnp.where(
        jnp.logical_and(invm > 0, jnp.expand_dims(state.is_active, axis=1)),
        space.gravity * jnp.ones_like(state.v.xy) + state.f.xy * invm,
        jnp.zeros_like(state.v.xy),
    )
    v_xy = state.v.xy + f_xy * space.dt
    f_ang = jnp.where(state.is_active, state.f.angle, 0.0)
    v_ang = state.v.angle + f_ang * shape.inv_moment() * space.dt
    v_xy = jnp.clip(
        v_xy * space.linear_damping,
        min=-space.max_velocity,
        max=space.max_velocity,
    )
    v_ang = jnp.clip(
        v_ang * space.angular_damping,
        min=-space.max_angular_velocity,
        max=space.max_angular_velocity,
    )
    # Damping: dv/dt + vc = 0 -> v(t) = v0 * exp(-tc)
    # v(t + dt) = v0 * exp(-tc - dtc) = v0 * exp(-tc) * exp(-dtc) = v(t)exp(-dtc)
    # Thus, linear/angular damping factors are actually exp(-dtc)
    return replace(state, v=Velocity(angle=v_ang, xy=v_xy), f=state.f.zeros_like())


def update_position(space: Space, state: State) -> State:
    v_dt = state.v * space.dt
    xy = state.p.xy + v_dt.xy
    angle = (state.p.angle + v_dt.angle + TWO_PI) % TWO_PI
    return replace(state, p=Position(angle=angle, xy=xy))


def init_contact_helper(
    space: Space,
    contact: Contact,
    a: Shape,
    b: Shape,
    p1: Position,
    p2: Position,
    v1: Velocity,
    v2: Velocity,
) -> ContactHelper:
    r1 = contact.pos - p1.xy
    r2 = contact.pos - p2.xy

    inv_mass1, inv_mass2 = a.inv_mass(), b.inv_mass()
    inv_moment1, inv_moment2 = a.inv_moment(), b.inv_moment()
    kn1 = _effective_mass(inv_mass1, inv_moment1, r1, contact.normal)
    kn2 = _effective_mass(inv_mass2, inv_moment2, r2, contact.normal)
    nx, ny = _get_xy(contact.normal)
    tangent = jnp.stack((-ny, nx), axis=-1)
    kt1 = _effective_mass(inv_mass1, inv_moment1, r1, tangent)
    kt2 = _effective_mass(inv_mass2, inv_moment2, r2, tangent)
    clipped_p = jnp.clip(space.allowed_penetration - contact.penetration, max=0.0)
    v_bias = -space.bias_factor / space.dt * clipped_p
    # k_normal, k_tangent, and v_bias should have (N(N-1)/2, N_contacts) shape
    chex.assert_equal_shape((contact.friction, kn1, kn2, kt1, kt2, v_bias))
    # Compute elasiticity * relative_vel
    dv = v2.rv(r2) - v1.rv(r1)
    vn = _vmap_dot(dv, contact.normal)
    return ContactHelper(  # type: ignore
        tangent=tangent,
        mass_normal=1 / (kn1 + kn2),
        mass_tangent=1 / (kt1 + kt2),
        v_bias=v_bias,
        bounce=vn * contact.elasticity,
        r1=r1,
        r2=r2,
        inv_mass1=inv_mass1,
        inv_mass2=inv_mass2,
        inv_moment1=inv_moment1,
        inv_moment2=inv_moment2,
        local_anchor1=p1.inv_rotate(r1),
        local_anchor2=p2.inv_rotate(r2),
        allow_bounce=vn <= -space.bounce_threshold,
    )


@chex.dataclass
class VelocitySolver:
    v1: jax.Array
    v2: jax.Array
    pn: jax.Array
    pt: jax.Array
    contact: jax.Array

    def update(self, new_contact: jax.Array) -> Self:
        continuing_contact = jnp.logical_and(self.contact, new_contact)
        pn = jnp.where(continuing_contact, self.pn, 0.0)
        pt = jnp.where(continuing_contact, self.pt, 0.0)
        return replace(self, pn=pn, pt=pt, contact=new_contact)


@jax.vmap
def apply_initial_impulse(
    contact: Contact,
    helper: ContactHelper,
    solver: VelocitySolver,
) -> VelocitySolver:
    """Warm starting by applying initial impulse"""
    p = helper.tangent * solver.pt + contact.normal * solver.pn
    v1 = solver.v1 - _axy(
        angle=helper.inv_moment1 * jnp.cross(helper.r1, p),
        xy=p * helper.inv_mass1,
    )
    v2 = solver.v2 + _axy(
        angle=helper.inv_moment2 * jnp.cross(helper.r2, p),
        xy=p * helper.inv_mass2,
    )
    return replace(solver, v1=v1, v2=v2)


def _rv_a2b(a: jax.Array, ra: jax.Array, b: jax.Array, rb: jax.Array):
    return Velocity.from_axy(b).rv(rb) - Velocity.from_axy(a).rv(ra)


@jax.vmap
def apply_velocity_normal(
    contact: Contact,
    helper: ContactHelper,
    solver: VelocitySolver,
) -> VelocitySolver:
    """
    Apply velocity constraints to the solver.
    Suppose that each shape has (N_contact, 1) or (N_contact, 2).
    """
    # Relative veclocity (from shape2 to shape1)
    dv = _rv_a2b(solver.v1, helper.r1, solver.v2, helper.r2)
    vt = jnp.dot(dv, helper.tangent)
    dpt = -helper.mass_tangent * vt
    # Clamp friction impulse
    max_pt = contact.friction * solver.pn
    pt = jnp.clip(solver.pt + dpt, min=-max_pt, max=max_pt)
    dpt_clamped = helper.tangent * (pt - solver.pt)
    # Velocity update by contact tangent
    dvt1 = _axy(
        angle=-helper.inv_moment1 * jnp.cross(helper.r1, dpt_clamped),
        xy=-dpt_clamped * helper.inv_mass1,
    )
    dvt2 = _axy(
        angle=helper.inv_moment2 * jnp.cross(helper.r2, dpt_clamped),
        xy=dpt_clamped * helper.inv_mass2,
    )
    # Compute Relative velocity again
    dv = _rv_a2b(solver.v1 + dvt1, helper.r1, solver.v2 + dvt2, helper.r2)
    vn = _vmap_dot(dv, contact.normal)
    dpn = helper.mass_normal * (-vn + helper.v_bias)
    # Accumulate and clamp impulse
    pn = jnp.clip(solver.pn + dpn, min=0.0)
    dpn_clamped = contact.normal * (pn - solver.pn)
    # Velocity update by contact normal
    dvn1 = _axy(
        angle=-helper.inv_moment1 * jnp.cross(helper.r1, dpn_clamped),
        xy=-dpn_clamped * helper.inv_mass1,
    )
    dvn2 = _axy(
        angle=helper.inv_moment2 * jnp.cross(helper.r2, dpn_clamped),
        xy=dpn_clamped * helper.inv_mass2,
    )
    # Filter dv
    return VelocitySolver(
        v1=jnp.where(solver.contact, dvn1 + dvt1, 0.0),
        v2=jnp.where(solver.contact, dvn2 + dvt2, 0.0),
        pn=pn,
        pt=pt,
        contact=solver.contact,
    )


@jax.vmap
def apply_bounce(
    contact: Contact,
    helper: ContactHelper,
    solver: VelocitySolver,
) -> tuple[jax.Array, jax.Array]:
    """
    Apply bounce (resititution).
    Suppose that each shape has (N_contact, 1) or (N_contact, 2).
    """
    # Relative veclocity (from shape2 to shape1)
    dv = _rv_a2b(solver.v1, helper.r1, solver.v2, helper.r2)
    vn = jnp.dot(dv, contact.normal)
    pn = -helper.mass_normal * (vn + helper.bounce)
    dpn = contact.normal * pn
    # Velocity update by contact normal
    dv1 = _axy(
        angle=-helper.inv_moment1 * jnp.cross(helper.r1, dpn),
        xy=-dpn * helper.inv_mass1,
    )
    dv2 = _axy(
        angle=helper.inv_moment2 * jnp.cross(helper.r2, dpn),
        xy=dpn * helper.inv_mass2,
    )
    # Filter dv
    allow_bounce = jnp.logical_and(solver.contact, helper.allow_bounce)
    return jnp.where(allow_bounce, dv1, 0.0), jnp.where(allow_bounce, dv2, 0.0)


@chex.dataclass
class PositionSolver:
    p1: jax.Array
    p2: jax.Array
    contact: jax.Array
    min_separation: jax.Array


@functools.partial(jax.vmap, in_axes=(None, None, None, 0, 0, 0))
def correct_position(
    bias_factor: float | jax.Array,
    linear_slop: float | jax.Array,
    max_linear_correction: float | jax.Array,
    contact: Contact,
    helper: ContactHelper,
    solver: PositionSolver,
) -> PositionSolver:
    """
    Correct positions to remove penetration.
    Suppose that each shape in contact and helper has (N_contact, 1) or (N_contact, 2).
    p1 and p2 should have xy: (1, 2) angle (1, 1) shape
    """
    # (N_contact, 2)
    p1, p2 = Position.from_axy(solver.p1), Position.from_axy(solver.p2)
    r1 = p1.rotate(helper.local_anchor1)
    r2 = p2.rotate(helper.local_anchor2)
    ga2_ga1 = r2 - r1 + p2.xy - p1.xy
    separation = jnp.dot(ga2_ga1, contact.normal) - contact.penetration
    c = jnp.clip(
        bias_factor * (separation + linear_slop),
        min=-max_linear_correction,
        max=0.0,
    )
    kn1 = _effective_mass(helper.inv_mass1, helper.inv_moment1, r1, contact.normal)
    kn2 = _effective_mass(helper.inv_mass2, helper.inv_moment2, r2, contact.normal)
    k_normal = kn1 + kn2
    impulse = jnp.where(k_normal > 0.0, -c / k_normal, 0.0)
    pn = impulse * contact.normal
    dp1 = _axy(
        angle=-helper.inv_moment1 * jnp.cross(r1, pn),
        xy=-pn * helper.inv_mass1,
    )
    dp2 = _axy(
        angle=helper.inv_moment2 * jnp.cross(r2, pn),
        xy=pn * helper.inv_mass2,
    )
    min_sep = jnp.fmin(solver.min_separation, separation)
    # Filter p1/p2
    dp1 = jnp.where(solver.contact, dp1, 0.0)
    dp2 = jnp.where(solver.contact, dp2, 0.0)
    return replace(solver, p1=dp1, p2=dp2, min_separation=min_sep)


def solve_constraints(
    space: Space,
    solver: VelocitySolver,
    p: Position,
    v: Velocity,
    contact: Contact,
) -> tuple[Velocity, Position, VelocitySolver]:
    """Resolve collisions by Sequential Impulse method"""
    idx1, idx2 = space._ci_total.index1, space._ci_total.index2

    def gather(a: jax.Array, b: jax.Array, orig: jax.Array) -> jax.Array:
        return orig.at[idx1].add(a).at[idx2].add(b)

    p1, p2 = p.get_slice(idx1), p.get_slice(idx2)
    v1, v2 = v.get_slice(idx1), v.get_slice(idx2)
    helper = init_contact_helper(
        space,
        contact,
        space._ci_total.shape1,
        space._ci_total.shape2,
        p1,
        p2,
        v1,
        v2,
    )
    # Warm up the velocity solver
    solver = replace(solver, v1=v1.into_axy(), v2=v2.into_axy())
    solver = apply_initial_impulse(contact, helper, solver)

    def vstep(
        _: int,
        vs: tuple[jax.Array, VelocitySolver],
    ) -> tuple[jax.Array, VelocitySolver]:
        v_i, solver_i = vs
        solver_i1 = apply_velocity_normal(contact, helper, solver_i)
        v_i1 = gather(solver_i1.v1, solver_i1.v2, v_i)
        return v_i1, replace(solver_i1, v1=v_i1[idx1], v2=v_i1[idx2])

    v_axy, solver = jax.lax.fori_loop(
        0,
        space.n_velocity_iter,
        vstep,
        (v.into_axy(), solver),
    )
    bv1, bv2 = apply_bounce(contact, helper, solver)
    v_axy = gather(bv1, bv2, v_axy)

    def pstep(
        _: int,
        ps: tuple[jax.Array, PositionSolver],
    ) -> tuple[jax.Array, PositionSolver]:
        p_i, solver_i = ps
        solver_i1 = correct_position(
            space.bias_factor,
            space.linear_slop,
            space.max_linear_correction,
            contact,
            helper,
            solver_i,
        )
        p_i1 = gather(solver_i1.p1, solver_i1.p2, p_i)
        return p_i1, replace(solver_i1, p1=p_i1[idx1], p2=p_i1[idx2])

    pos_solver = PositionSolver(
        p1=p1.into_axy(),
        p2=p2.into_axy(),
        contact=solver.contact,
        min_separation=jnp.zeros_like(p1.angle),
    )
    p_axy, pos_solver = jax.lax.fori_loop(
        0,
        space.n_position_iter,
        pstep,
        (p.into_axy(), pos_solver),
    )
    return Velocity.from_axy(v_axy), Position.from_axy(p_axy), solver


def step(
    space: Space,
    stated: StateDict,
    solver: VelocitySolver,
) -> tuple[StateDict, VelocitySolver, Contact]:
    state = update_velocity(space, space.shaped.concat(), stated.concat())
    contact = space.check_contacts(stated.update(state))
    penetration_allowed = contact.penetration >= space.speculative_distance
    v, p, solver = solve_constraints(
        space,
        solver.update(penetration_allowed),
        state.p,
        state.v,
        contact,
    )
    state = update_position(space, replace(state, v=v, p=p))
    return stated.update(state), solver, contact



import chex
import jax
import jax.numpy as jnp
import pytest

from phyjax2d import Space, SpaceBuilder, Vec2d, circle_overlap


@pytest.fixture
def space() -> Space:
    builder = SpaceBuilder(gravity=(0.0, 0.0))

    builder.add_circle(radius=2.0)
    builder.add_circle(radius=1.0)
    builder.add_segment(p1=Vec2d(-4.0, 0.0), p2=Vec2d(4.0, 0.0))
    builder.add_polygon(points=[Vec2d(2.0, 0.0), Vec2d(2.0, 2.0), Vec2d(0.0, 0.0)])
    builder.add_polygon(points=[Vec2d(2.0, 0.0), Vec2d(0.0, 2.0), Vec2d(0.0, 0.0)])
    return builder.build()


def test_circle_overlap(space: Space) -> None:
    sd = space.zeros_state().nested_replace(
        "circle.p.xy",
        jnp.array([[3.0, 3.0], [0.0, 0.0]]),
    )
    sd = sd.nested_replace("segment.p.xy", jnp.array([[0.0, -1.0]]))
    sd = sd.nested_replace("triangle.p.xy", jnp.array([[0.0, -5.0], [-2.0, -5.0]]))
    ol1 = circle_overlap(space.shaped, sd, jnp.array([0.0, 3.0]), 1.4)
    assert ol1.item()
    ol2 = circle_overlap(space.shaped, sd, jnp.array([0.0, 3.0]), 0.9)
    assert not ol2.item()
    ol3 = circle_overlap(space.shaped, sd, jnp.array([0.0, -4.0]), 1.0)
    assert ol3.item()
    ol3 = circle_overlap(space.shaped, sd, jnp.array([0.0, -4.0]), 1.0)
    assert ol3.item()
    ol4 = circle_overlap(space.shaped, sd, jnp.array([1.0, -6.0]), 1.1)
    assert ol4.item()
    ol5 = circle_overlap(space.shaped, sd, jnp.array([4.0, -5.0]), 1.8)
    assert not ol5.item()
    ol6 = circle_overlap(space.shaped, sd, jnp.array([-2.0, -3.0]), 0.1)
    assert ol6.item()
    ol7 = circle_overlap(space.shaped, sd, jnp.array([-2.0, 0.0]), 1.1)
    assert ol7.item()



from phyjax2d.impl import (
    Capsule,
    Circle,
    Force,
    Polygon,
    Position,
    Segment,
    ShapeDict,
    Space,
    State,
    StateDict,
    Velocity,
    VelocitySolver,
    empty,
    get_relative_angle,
    step,
)
from phyjax2d.raycast import (
    Raycast,
    circle_raycast,
    segment_raycast,
    thin_polygon_raycast,
)
from phyjax2d.utils import (
    Color,
    SpaceBuilder,
    circle_overlap,
    make_approx_circle,
    make_square_segments,
)
from phyjax2d.vec2d import Vec2d

__version__ = "0.3.1"



import jax
import jax.numpy as jnp
import pytest

from phyjax2d import Space, SpaceBuilder, Vec2d, step


@pytest.fixture
def space() -> Space:
    builder = SpaceBuilder(gravity=(0.0, -9.8))

    builder.add_circle(radius=2.0)
    builder.add_circle(radius=4.0)
    builder.add_segment(p1=Vec2d(-10.0, 0.0), p2=Vec2d(10.0, 0.0))
    return builder.build()


def test_circle_fall(space: Space) -> None:
    sd = space.zeros_state().nested_replace(
        "circle.p.xy",
        jnp.array([[-4.0, 6.0], [6.0, 8.0]]),
    )
    vs = space.init_solver()
    nsd, _, _ = step(space, sd, vs)
    assert nsd.circle.p.xy[0][1] < 6.0
    assert nsd.circle.p.xy[1][1] < 8.0


def test_circle_fall_jit(space: Space) -> None:
    sd = space.zeros_state().nested_replace(
        "circle.p.xy",
        jnp.array([[-4.0, 6.0], [6.0, 8.0]]),
    )
    vs = space.init_solver()
    jit_step = jax.jit(step, static_argnums=(0,))
    for _ in range(10):
        sd, _, _ = jit_step(space, sd, vs)
    assert sd.circle.p.xy[0][1] > 0.0
    assert sd.circle.p.xy[1][1] > 0.0
